#include "endpointinc.h"

#include "dh.h"
#include "xmlgeninc/_xmlgen_.hpp"
#include "erroroccured.h"
#include "variantdef.h"
#include "viewcontextimpl.h"
#include "scriptexchange.h"

namespace limax {
	namespace helper {

#ifdef LIMAX_VS_DEBUG_MEMORY_LEAKS_DETECT

		struct OsSystemInit
		{
			OsSystemInit()
			{
				_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
			}

			void Startup();
			void Cleanup();
		};
#else
		struct OsSystemInit
		{
			OsSystemInit() {}

			void Startup();
			void Cleanup();
		};

#endif

#ifdef LIMAX_OS_WINDOWS
		void OsSystemInit::Startup()
		{
			WSADATA data;
			WSAStartup(MAKEWORD(2, 0), &data);
		}
		void OsSystemInit::Cleanup()
		{
			::WSACleanup();
		}
#else
		void OsSystemInit::Startup() {}
		void OsSystemInit::Cleanup() {}
#endif

		static OsSystemInit ossysteminit;

	} // namespace helper {
} // namespace limax {

namespace limax {

	EndpointListener::EndpointListener() {}
	EndpointListener::~EndpointListener() {}

	EndpointManager::EndpointManager() {}
	EndpointManager::~EndpointManager() {}

	Transport::Transport() {}
	Transport::~Transport() {}

	EndpointConfig::EndpointConfig() {}
	EndpointConfig::~EndpointConfig() {}

	EndpointConfigBuilder::EndpointConfigBuilder() {}
	EndpointConfigBuilder::~EndpointConfigBuilder() {}
	namespace helper {

		struct EndpointConfigData
		{
			std::string serverIp;
			int serverPort;
			int dhGroup;
			std::string username;
			std::string token;
			std::string platflag;
			bool ispingonly;
			std::shared_ptr<State> state;
			std::vector< std::shared_ptr<View::ViewCreatorManager> > vcms;
			std::vector<int16_t> vpvids;
			std::shared_ptr<ScriptEngineHandle> script;
			Executor executor;
		};

		struct EndpointConfigImpl : public EndpointConfig
		{
			EndpointConfigData	data;
			EndpointConfigImpl(const EndpointConfigData& _data) : data(_data){}
			virtual ~EndpointConfigImpl() {}
			virtual const std::string& getServerIP() const override
			{
				return data.serverIp;
			}
			virtual int getServerPort() const override
			{
				return data.serverPort;
			}
			virtual int getDHGroup() const override
			{
				return data.dhGroup;
			}
			virtual const std::string& getUserName() const override
			{
				return data.username;
			}
			virtual const std::string& getToken() const override
			{
				return data.token;
			}
			virtual const std::string& getPlatFlag()const  override
			{
				return data.platflag;
			}
			virtual bool isPingServerOnly() const override
			{
				return data.ispingonly;
			}
			virtual std::shared_ptr<State> getEndpointState() const override
			{
				return data.state;
			}
			virtual const std::vector< std::shared_ptr<View::ViewCreatorManager> >& getStaticViewCreatorManagers() const override
			{
				return data.vcms;
			}
			virtual const std::vector<int16_t>& getVariantProviderIds() const override
			{
				return data.vpvids;
			}
			virtual std::shared_ptr<ScriptEngineHandle> getScriptEngineHandle() const override
			{
				return data.script;
			}
			virtual Executor getExecutor() const override
			{
				return data.executor;
			}
		};

		struct EndpointConfigBuilderImpl : public EndpointConfigBuilder
		{
			EndpointConfigData data;
			std::weak_ptr<EndpointConfigBuilder> instance;
			EndpointConfigBuilderImpl(const std::string& serverip, int serverport, bool pingonly)
			{
				data.ispingonly = pingonly;
				data.serverIp = serverip;
				data.serverPort = serverport;
				data.dhGroup = 2;
				data.executor = [](Runnable r){ r(); };
				data.state = std::shared_ptr<State>(new State(limax::getEndpointStateEndpointClient()));
			}
			virtual ~EndpointConfigBuilderImpl() { }
			virtual std::shared_ptr<EndpointConfigBuilder> endpointState(std::initializer_list<std::shared_ptr<State>> states) override
			{
				data.state = std::shared_ptr<State>(new State(limax::getEndpointStateEndpointClient()));
				for (auto& i : states)
					data.state->addStub(i);
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> endpointState(std::vector<std::shared_ptr<State>> states) override
			{
				data.state = std::shared_ptr<State>(new State(limax::getEndpointStateEndpointClient()));
				for (auto& i : states)
					data.state->addStub(i);
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> staticViewCreatorManagers(std::initializer_list<std::shared_ptr<View::ViewCreatorManager>> vcms) override
			{
				data.vcms.clear();
				data.vcms.reserve(vcms.size());
				data.vcms.insert(data.vcms.end(), vcms.begin(), vcms.end());
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> staticViewCreatorManagers(std::vector<std::shared_ptr<View::ViewCreatorManager>> vcms) override
			{
				data.vcms.clear();
				data.vcms.reserve(vcms.size());
				data.vcms.insert(data.vcms.end(), vcms.begin(), vcms.end());
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> variantProviderIds(std::initializer_list<int16_t> pvids) override
			{
				data.vpvids.clear();
				data.vpvids.reserve(pvids.size());
				data.vpvids.insert(data.vpvids.end(), pvids.begin(), pvids.end());
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> variantProviderIds(std::vector<int16_t> pvids) override
			{
				data.vpvids.clear();
				data.vpvids.reserve(pvids.size());
				data.vpvids.insert(data.vpvids.end(), pvids.begin(), pvids.end());
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> scriptEngineHandle(std::shared_ptr<ScriptEngineHandle> handle) override
			{
				data.script = handle;
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfigBuilder> executor(Executor executor) override
			{
				data.executor = executor;
				return instance.lock();
			}
			virtual std::shared_ptr<EndpointConfig> build() const override
			{
				return std::shared_ptr<EndpointConfig>(new EndpointConfigImpl(data));
			}
		};

		static inline void mapPvidsAppendValue(hashmap<int16_t, int8_t>& pvids, int16_t s, int nv)
		{
			pvids[s] = (int8_t)(nv | pvids[s]);
		}
		static inline void mapPvidsAppendValue(hashmap<int16_t, int8_t>& pvids, const hashset<int16_t>& set, int nv)
		{
			for (auto& it : set)
				mapPvidsAppendValue(pvids, it, nv);
		}
		static inline const hashset<int16_t> makeProtocolPvidSet(std::shared_ptr<EndpointConfig> config)
		{
			return config->getEndpointState()->getProviderIds();
		}

		static inline const hashset<int16_t> makeStaticPvidSet(std::shared_ptr<EndpointConfig> config)
		{
			hashset<int16_t> pvidset;
			const auto& vcms = config->getStaticViewCreatorManagers();
			for (auto& vcm : vcms)
				pvidset.insert(vcm->getProviderId());
			return pvidset;
		}

		static inline const hashset<int16_t> makeVariantPvidSet(std::shared_ptr<EndpointConfig> config)
		{
			hashset<int16_t> pvidset;
			const auto& pvids = config->getVariantProviderIds();
			pvidset.insert(pvids.begin(), pvids.end());
			return pvidset;
		}

		static inline const hashset<int16_t> makeScriptPvidSet(std::shared_ptr<EndpointConfig> config)
		{
			hashset<int16_t> pvidset;
			if (auto handle = config->getScriptEngineHandle())
				pvidset.insert(handle->getProviders().begin(), handle->getProviders().end());
			return pvidset;
		}

		static inline const hashset<int16_t> makeAuanyPvidSet()
		{
			hashset<int16_t> pvidset;
			pvidset.insert(getAuanyviewsViewCreatorManager()->getProviderId());
			return pvidset;
		}

		struct MakeProviderMapException
		{
			MakeProviderMapException(int code) : errorcode(code){}
			MakeProviderMapException(int code, const std::string& msg) : errorcode(code), message(msg){}
			int errorcode;
			std::string message;
		};

		static inline void makeProviderMap(hashmap<int16_t, int8_t>& pvids, std::shared_ptr<EndpointConfig> config)
		{
			mapPvidsAppendValue(pvids, makeProtocolPvidSet(config), limax::defines::SessionType::ST_PROTOCOL);
			mapPvidsAppendValue(pvids, makeStaticPvidSet(config), limax::defines::SessionType::ST_STATIC);
			mapPvidsAppendValue(pvids, makeVariantPvidSet(config), limax::defines::SessionType::ST_VARIANT);
			mapPvidsAppendValue(pvids, makeScriptPvidSet(config), limax::defines::SessionType::ST_SCRIPT);
			if (pvids.empty())
				throw MakeProviderMapException(SYSTEM_ENDPOINT_NO_AVAILABLE_PROVIDER);
			mapPvidsAppendValue(pvids, makeAuanyPvidSet(), limax::defines::SessionType::ST_STATIC);
		}

		struct ViewContextMap
		{
			struct ViewContextTypeHash
			{
				size_t operator()(ViewContext::Type t) const
				{
					return (size_t)t;
				}
			};
			struct ViewContextCollection
			{
				std::unordered_map<ViewContext::Type, std::shared_ptr<AbstractViewContext>, ViewContextTypeHash> map;
				void put(ViewContext::Type type, std::shared_ptr<AbstractViewContext> vc)
				{
					map.insert(std::make_pair(type, vc));
				}
				void onSyncViewToClients(limax::endpoint::providerendpoint::SyncViewToClients* protocol)
				{
					for (auto& it : map)
						it.second->onSyncViewToClients(protocol);
				}
				void clear()
				{
					for (auto& it : map)
						it.second->clear();
					map.clear();
				}
				std::shared_ptr<AbstractViewContext> getViewContext(ViewContext::Type type)
				{
					auto it = map.find(type);
					return it == map.end() ? nullptr : it->second;
				}
				int getSize() const
				{
					return (int)map.size();
				}
				std::shared_ptr<AbstractViewContext> getViewContext(ViewContext::Type type) const
				{
					auto it = map.find(type);
					return it == map.end() ? nullptr : it->second;
				}
			};
			hashmap<int16_t, ViewContextCollection> map;
			void put(ViewContext::Type type, int16_t pvid, std::shared_ptr<AbstractViewContext> vc)
			{
				ViewContextCollection& cc = map[pvid];
				cc.put(type, vc);
			}
			void onSyncViewToClients(limax::endpoint::providerendpoint::SyncViewToClients* protocol)
			{
				auto it = map.find(protocol->providerid);
				if (it != map.end())
					it->second.onSyncViewToClients(protocol);
			}
			void clear()
			{
				for (auto& it : map)
					it.second.clear();
				map.clear();
			}
			const std::shared_ptr<ViewContext> getViewContext(int16_t pvid, ViewContext::Type type) const
			{
				auto it = map.find(pvid);
				if (it == map.end())
					return nullptr;
				else
					return it->second.getViewContext(type);
			}
		};

		static std::atomic<EndpointManager*> g_defaultEndpointManager;

	} // namespace helper {

	std::shared_ptr<EndpointConfigBuilder> Endpoint::createLoginConfigBuilder(const std::string& serverip, int serverport, const std::string& username, const std::string& token, const std::string& platflag)
	{
		auto cb = std::shared_ptr<helper::EndpointConfigBuilderImpl>(new helper::EndpointConfigBuilderImpl(serverip, serverport, false));
		cb->instance = cb;
		cb->data.username = username;
		cb->data.token = token;
		cb->data.platflag = platflag;
		return cb;
	}

	std::shared_ptr<EndpointConfigBuilder> Endpoint::createPingOnlyConfigBuilder(const std::string& serverip, int serverport)
	{
		auto cb = std::shared_ptr<helper::EndpointConfigBuilderImpl>(new helper::EndpointConfigBuilderImpl(serverip, serverport, true));
		cb->instance = cb;
		cb->data.ispingonly = true;
		return cb;
	}

	class EndpointImpl : public EndpointManager, public Transport, public NetSession::Listener
	{
		std::shared_ptr<EndpointConfig>  m_config;
		std::shared_ptr<EndpointListener> m_listener;
		NetSession*	m_session = nullptr;
		std::recursive_mutex	m_session_mutex;
		int64_t						m_sessionid = -1;
		int64_t						m_accountFlags = 0;
		struct sockaddr_in					m_localaddress;
		struct sockaddr_in					m_peeraddress;
		void*	m_object = nullptr;
		hashmap<int16_t, int8_t>	m_pvids;
		std::shared_ptr<limax::DHContext>	dhcontext;
		Dispatcher dispatcher;
		DelayedRunnable ping_timeout_runnable;
		DelayedRunnable ping_alive_delay_runnable;
		volatile bool timeout = false;
		helper::ViewContextMap	m_viewContextMap;
		helper::ScriptExchange	m_scriptexchange;
		volatile enum LoginStatus
		{
			LOGINING, LOGINED_NOTIFY, LOGINED_DONE, RETRY,
		} loginstatus = LoginStatus::LOGINING;
		bool m_canretrylogin = true;
	private:
		void startPingAndKeepAlive()
		{
			sendProtocol(limax::endpoint::switcherendpoint::PingAndKeepAlive(std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - std::chrono::steady_clock::time_point::min()).count()));
			ping_timeout_runnable = DelayedRunnable([this](){
				timeout = true;
				close();
			}, 500, 10);
			Engine::execute(ping_timeout_runnable);
		}
		void startConnect()
		{
			NetSession::create(this, this, m_config->getServerIP(), m_config->getServerPort());
		}
		void executeSessionTask(std::function<void(NetSession*&)> r)
		{
			std::lock_guard<std::recursive_mutex> l(m_session_mutex);
			if (m_session)
				r(m_session);
		}
		void createSessionTask(std::function<void(NetSession*&)> r)
		{
			std::lock_guard<std::recursive_mutex> l(m_session_mutex);
			r(m_session);
		}
		void onAddSession(NetSession* _session, const struct sockaddr_in& local, const struct sockaddr_in& peer) override
		{
			createSessionTask([=](NetSession*& session){
				session = _session;
				session->setState(getEndpointStateEndpointClient());
				session->setInputSecurity(false);
				session->setOutputSecurity(false);
				m_localaddress = local;
				m_peeraddress = peer;
				dispatcher([this](){m_listener->onSocketConnected(); });
				if (m_config->isPingServerOnly())
				{
					startPingAndKeepAlive();
				}
				else
				{
					const int dh_group = m_config->getDHGroup();
					dhcontext = createDHContext(dh_group);
					const std::vector<unsigned char>& data = dhcontext->generateDHResponse();
					session->send(limax::endpoint::switcherendpoint::CHandShake(dh_group, Octets(&data[0], data.size())));
				}
			});
		}
		void onAbortSession(NetSession* session) override
		{
			dispatcher([this]()
			{
				m_listener->onAbort(this);
				m_listener->onManagerUninitialized(this);
			});
			dispatcher.await();
			Engine::remove(this);
		}
		void onDelSession(NetSession* session) override
		{
			executeSessionTask([](NetSession*& session){ session = nullptr; });
			ping_timeout_runnable.cancel();
			ping_alive_delay_runnable.cancel();
			if (timeout)
				pushErrorOccured(SOURCE_LIMAX, ENDPOINT_PING_TIMEOUT, "ping server time out");
			dispatcher.await();
			dispatcher([this, session]()
			{
				if (LoginStatus::RETRY == loginstatus)
				{
					startConnect();
				}
				else
				{
					if (LoginStatus::LOGINING == loginstatus)
					{
						if (!m_config->isPingServerOnly())
							m_listener->onAbort(this);
					}
					else
					{
						if (LoginStatus::LOGINED_DONE == loginstatus)
						{
							m_listener->onTransportRemoved(this);
						}
						if (LoginStatus::LOGINED_DONE == loginstatus || LoginStatus::LOGINED_NOTIFY == loginstatus) {
							EndpointManager* tmp = this;
							helper::g_defaultEndpointManager.compare_exchange_weak(tmp, nullptr);
							m_viewContextMap.clear();
							m_scriptexchange.onUnload([this](int source, int code, const std::string& message){ fireErrorOccured(source, code, message); });
						}
					}
					m_listener->onManagerUninitialized(this);
					Engine::execute([this](){
						dispatcher.await();
						ping_timeout_runnable.join();
						ping_alive_delay_runnable.join();
						Engine::remove(this);
					});
				}
			});
		}
		void onProtocol(Protocol* p) override
		{
			dispatcher([p](){
				p->process();
				p->destroy();
			});
		}
		void onCheckUnknownProtocol(NetSession* session, Protocol::Type type, int size) override
		{
			std::stringstream ss;
			ss << "type = " << type << " size = " << size;
			pushErrorOccured(SOURCE_ENDPOINT, SYSTEM_ENDPOINT_RECV_UNKNOWN_PROTOCOL, ss.str());
			throw Protocol::Exception();
		}
		void createStaticViewContexts()
		{
			for (auto& vcm : m_config->getStaticViewCreatorManagers())
			{
				auto pvid = vcm->getProviderId();
				auto vc = helper::StaticViewContextImpl::create(this, pvid, vcm);
				m_viewContextMap.put(ViewContext::Static, pvid, vc);
			}

			{
				auto pvid = getAuanyviewsViewCreatorManager()->getProviderId();
				auto vc = helper::StaticViewContextImpl::create(this, pvid, getAuanyviewsViewCreatorManager());
				m_viewContextMap.put(ViewContext::Static, pvid, vc);
			}
		}
		void createVariantViewContexts(const hashmap<int16_t, limax::defines::VariantDefines>& vdmap)
		{
			for (auto& it : vdmap)
			{
				auto pvid = it.first;
				const auto& vd = it.second;
				auto vc = helper::VariantViewContextImpl::create(this, pvid, vd);
				if (vc)
					m_viewContextMap.put(ViewContext::Variant, pvid, vc);
				else
					close();
			}
		}
		void createScriptViewContexts(const std::string& defines)
		{
			m_scriptexchange.onLoad(this, m_config->getScriptEngineHandle(), defines, [this](int source, int code, const std::string& message){ fireErrorOccured(source, code, message); });
		}
	public:
		EndpointImpl(std::shared_ptr<EndpointConfig> config, const hashmap<int16_t, int8_t>& pvids, EndpointListener* listener)
			: m_config(config), m_listener(std::shared_ptr<EndpointListener>(listener, [](EndpointListener *listener){ listener->destroy(); }))
			, m_pvids(pvids), dispatcher(config->getExecutor())
		{
			Engine::add(this);
			dispatcher([this](){ m_listener->onManagerInitialized(this, m_config.get()); });
			if (config->isPingServerOnly())
				m_canretrylogin = false;
			startConnect();
		}
		void close() override
		{
			executeSessionTask([](NetSession*& session) { session->close(); });
		}
		virtual int64_t getSessionId() const override{ return m_sessionid; }
		virtual int64_t getAccountFlags() const override{ return m_accountFlags; }
		virtual void sendProtocol(const limax::Protocol& p) override
		{
			executeSessionTask([&p](NetSession*& session) { session->send(&p); });
		}
		virtual ViewContext* getViewContext(int16_t pvid, ViewContext::Type type) const override
		{
			return m_viewContextMap.getViewContext(pvid, type).get();
		}
		void unmarshalViewException(int16_t pvid, int16_t classIndex, int32_t instanceIndex)
		{
			std::stringstream ss;
			ss << "providerId = " << pvid << " classIndex = " << classIndex << " instanceIndex = " << instanceIndex;
			fireErrorOccured(SOURCE_ENDPOINT, SYSTEM_VIEW_MARSHAL_EXCEPTION, ss.str());
			close();
		}
		void fireErrorOccured(int source, int code, const std::string& info)
		{
			m_listener->onErrorOccured(source, code, info);
		}
		void pushErrorOccured(int source, int code, const std::string& info)
		{
			dispatcher([=](){ fireErrorOccured(source, code, info); });
		}
		void onProtocolOnlineAnnounce(limax::endpoint::switcherendpoint::OnlineAnnounce* p)
		{
			if (SOURCE_LIMAX == p->errorSource && SUCCEED == p->errorCode)
			{
				m_accountFlags = p->flags;
				m_sessionid = p->sid;
				createStaticViewContexts();
				createVariantViewContexts(p->variantdefines);
				createScriptViewContexts(p->scriptdefines);
				loginstatus = LoginStatus::LOGINED_NOTIFY;
				EndpointManager*tmp = nullptr;
				helper::g_defaultEndpointManager.compare_exchange_weak(tmp, this);
				m_listener->onTransportAdded(this);
				loginstatus = LoginStatus::LOGINED_DONE;
				startPingAndKeepAlive();
			}
			else if (m_canretrylogin && SOURCE_LIMAX == p->errorSource && SWITCHER_SESSION_LOGINED == p->errorCode)
			{
				m_canretrylogin = false;
				loginstatus = LoginStatus::RETRY;
			}
			else
			{
				fireErrorOccured(p->errorSource, p->errorCode, "switcherendpoint::OnlineAnnounce");
			}
		}
		void onProtocolSKeyExchange(limax::endpoint::switcherendpoint::SHandShake* p)
		{
			executeSessionTask([this, p](NetSession *&session) {
				const std::vector<unsigned char> keydata = dhcontext->computeDHKey((unsigned char*)p->dh_data.begin(), (int32_t)p->dh_data.size());
				int32_t half = (int32_t)keydata.size() / 2;
				{
					HmacMD5 hmac(limax::Codec::Null(), (int8_t*)&m_peeraddress.sin_addr.s_addr, 0, sizeof(m_peeraddress.sin_addr.s_addr));
					hmac.update((int8_t*)&keydata[0], 0, half);
					session->setOutputSecurity(p->c2sneedcompress, Octets(hmac.digest(), 16));
				}
				{
					HmacMD5 hmac(limax::Codec::Null(), (int8_t*)&m_peeraddress.sin_addr.s_addr, 0, sizeof(m_peeraddress.sin_addr.s_addr));
					hmac.update((int8_t*)&keydata[0], half, (int32_t)keydata.size() - half);
					session->setInputSecurity(p->s2cneedcompress, Octets(hmac.digest(), 16));
				}
				m_listener->onKeyExchangeDone();
				dhcontext.reset();
				limax::endpoint::switcherendpoint::SessionLoginByToken protocol;
				protocol.username = m_config->getUserName();
				protocol.token = m_config->getToken();
				protocol.platflag = m_config->getPlatFlag();
				protocol.report_ip.insert(protocol.report_ip.begin(), &m_localaddress.sin_addr.s_addr, sizeof(m_localaddress.sin_addr.s_addr));
				protocol.report_port = ntohs(m_localaddress.sin_port);
				protocol.pvids.insert(m_pvids.begin(), m_pvids.end());
				session->send(&protocol);
				std::shared_ptr<State> state = std::shared_ptr<State>(new State);
				state->addStub(getEndpointStateEndpointClient());
				state->addStub(m_config->getEndpointState());
				session->setState(state);
			});
		}
		void onProtocolPingAndKeepAlive(limax::endpoint::switcherendpoint::PingAndKeepAlive* p)
		{
			ping_timeout_runnable.cancel();
			m_listener->onKeepAlived((int)(std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - std::chrono::steady_clock::time_point::min()).count() - p->timestamp));
			if (!m_config->isPingServerOnly())
				Engine::execute(ping_alive_delay_runnable = DelayedRunnable([this](){startPingAndKeepAlive(); }, 500, 100));
		}
		void onProtocolSessionKick(limax::endpoint::switcherendpoint::SessionKick* p)
		{
			fireErrorOccured(SOURCE_LIMAX, p->error, "switcherendpoint::SessionKick");
			m_scriptexchange.onClose(p->error);
		}
		void onProtocolSyncViewToClients(limax::endpoint::providerendpoint::SyncViewToClients* protocol)
		{
			m_viewContextMap.onSyncViewToClients(protocol);
			m_scriptexchange.onSyncViewToClients(protocol, [this](int source, int code, const std::string& message){ fireErrorOccured(source, code, message); });
		}
		EndpointManager* getManager() { return this; }
		const struct sockaddr_in& getPeerAddress() const { return m_peeraddress; }
		const struct sockaddr_in& getLocalAddress() const { return m_localaddress; }
		void* getSessionObject() { return m_object; }
		void setSessionObject(void *obj){ m_object = obj; }
		Transport* getTransport() { return this; }
	};

	namespace erroroccured {

		void fireErrorOccured(EndpointManager* manager, int type, int code, const std::string& info)
		{
			if (auto impl = dynamic_cast<EndpointImpl*>(manager))
				impl->fireErrorOccured(type, code, info);
		}

	} // namespace erroroccured {

	void Endpoint::openEngine(){ Engine::open(); }

	void Endpoint::closeEngine(Runnable done){ Engine::close(done); }

	void Endpoint::start(std::shared_ptr<EndpointConfig> config, EndpointListener* handler)
	{
		Engine::execute([config, handler](){
			hashmap<int16_t, int8_t> pvids;
			if (!config->isPingServerOnly())
			{
				try
				{
					helper::makeProviderMap(pvids, config);
				}
				catch (helper::MakeProviderMapException& e)
				{
					config->getExecutor()([=](){handler->onErrorOccured(SOURCE_ENDPOINT, e.errorcode, e.message); });
					return;
				}
			}
			new EndpointImpl(config, pvids, handler);
		});
	}

	EndpointManager* Endpoint::getDefaultEndpointManager()
	{
		return helper::g_defaultEndpointManager.load();
	}

	namespace endpoint {
		namespace switcherendpoint {

			void SHandShake::process()
			{
				if (EndpointImpl* manager = dynamic_cast<EndpointImpl*>(getTransport()->getManager()))
					manager->onProtocolSKeyExchange(this);
			}

			void OnlineAnnounce::process()
			{
				if (EndpointImpl* manager = dynamic_cast<EndpointImpl*>(getTransport()->getManager()))
					manager->onProtocolOnlineAnnounce(this);
			}

			void PingAndKeepAlive::process()
			{
				if (EndpointImpl* manager = dynamic_cast<EndpointImpl*>(getTransport()->getManager()))
					manager->onProtocolPingAndKeepAlive(this);
			}

			void SessionKick::process()
			{
				if (EndpointImpl* manager = dynamic_cast<EndpointImpl*>(getTransport()->getManager()))
					manager->onProtocolSessionKick(this);
			}

			void PortForward::process() {}
			void CHandShake::process() {}
			void SessionLoginByToken::process() {}

		} // namespace switcherendpoint { 

		namespace providerendpoint {

			void SyncViewToClients::process()
			{
				auto impl = dynamic_cast<EndpointImpl*>(getTransport()->getManager());
				if (impl)
				{
					try
					{
						impl->onProtocolSyncViewToClients(this);
					}
					catch (MarshalException&)
					{
						impl->unmarshalViewException(providerid, classindex, instanceindex);
					}
				}
			}

			void SendControlToServer::process() {}

		} // namespace providerendpoint { 
	} // namespace endpoint {

	namespace helper { void ExpansionImplCleanup(); }

	std::mutex Engine::mutex;
	std::condition_variable_any Engine::cond;
	std::unordered_set<EndpointImpl*> Engine::set;
	ThreadPool *Engine::pool;
	void Engine::open()
	{
		std::lock_guard<std::mutex> l(mutex);
		if (!pool)
		{
			helper::ossysteminit.Startup();
			pool = new ThreadPool(30000);
		}
	}

	void Engine::close(Runnable done)
	{
		helper::ExpansionImplCleanup();
		std::thread([done](){
			std::lock_guard<std::mutex> l(mutex);
			if (pool)
			{
				for (auto& e : std::vector<EndpointImpl*>(set.begin(), set.end()))
					e->close();
				while (set.size())
					cond.wait(mutex);
				delete pool;
				pool = nullptr;
				helper::ossysteminit.Cleanup();
			}
			if (done)
				done();
		}).detach();
	}

	void Engine::add(EndpointImpl* e)
	{
		std::lock_guard<std::mutex> l(mutex);
		set.insert(e);
	}

	void Engine::remove(EndpointImpl* e)
	{
		std::lock_guard<std::mutex> l(mutex);
		if (set.erase(e) == 0)
			return;
		if (set.empty())
			cond.notify_one();
		delete e;
	}

} // namespace limax {

namespace limax {
	namespace helper {

		static class ExpansionImpl
		{
			struct ExperienceUpgradeParam
			{
				EndpointManager* manager;
				std::string username;
				std::string token;
				std::string platflag;
				std::function<void(int, int)> onresult;
				DelayedRunnable timeout;
			};

			std::mutex	locker;
			std::atomic<int64_t> experienceUpgradeSnGenerator;
			hashmap<int64_t, ExperienceUpgradeParam> experienceUpgradeRequestMap;
		public:
			void experienceUpgrade(EndpointManager* manager, const std::string& username, const std::string& token, const std::string& platflag, std::function<void(int, int)> onresult)
			{
				ExperienceUpgradeParam param;

				param.manager = manager;
				param.username = username;
				param.token = token;
				param.platflag = platflag;
				param.onresult = onresult;

				int64_t sn = experienceUpgradeSnGenerator++;
				param.timeout = DelayedRunnable([this, sn](){

					ExperienceUpgradeParam param;
					{
						std::lock_guard<std::mutex> __lg__(locker);
						auto it = experienceUpgradeRequestMap.find(sn);
						if (it == experienceUpgradeRequestMap.end())
							return;
						param = it->second;
						experienceUpgradeRequestMap.erase(it);
					}
					param.onresult(defines::ErrorSource::LIMAX, defines::ErrorCodes::AUANY_EXPERIENCE_UPGRADE_CLIENT_TIMEOUT);
				}, 500, 120);
				{
					std::lock_guard<std::mutex> __lg__(locker);
					experienceUpgradeRequestMap.insert(std::make_pair(sn, param));
				}
				Engine::execute(param.timeout);
				endpoint::auanyviews::Service::getInstance(manager)->ExperienceUpgrade(sn);
			}

			void onViewOpen(endpoint::auanyviews::ExperienceUpgrade* view)
			{
				view->registerListener("sn", [this, view](const ViewChangedEvent& e)
				{
					int64_t sn = *(int64_t*)e.getValue();
					ExperienceUpgradeParam param;
					{
						std::lock_guard<std::mutex> __lg__(locker);
						auto it = experienceUpgradeRequestMap.find(sn);
						if (it == experienceUpgradeRequestMap.end())
							return;
						param = it->second;
					};
					view->Upgrade(param.username, param.token, param.platflag);
				});

				view->registerListener("errorcodes", [this, view](const ViewChangedEvent& e)
				{
					const auto& errorCodes = *(auanyviews::ErrorCodes*)e.getValue();
					view->visitSn([this, errorCodes](int64_t sn)
					{
						ExperienceUpgradeParam param;
						{
							std::lock_guard<std::mutex> __lg__(locker);
							auto it = experienceUpgradeRequestMap.find(sn);
							if (it == experienceUpgradeRequestMap.end())
								return;
							param = it->second;
							experienceUpgradeRequestMap.erase(it);
						};
						param.timeout.cancel();
						param.onresult(errorCodes.errorSource, errorCodes.errorCode);
					});
				});
			}

			void onViewClose(endpoint::auanyviews::ExperienceUpgrade* view)
			{
				view->visitSn([this](int64_t sn)
				{
					ExperienceUpgradeParam param;
					{
						std::lock_guard<std::mutex> __lg__(locker);
						auto it = experienceUpgradeRequestMap.find(sn);
						if (it == experienceUpgradeRequestMap.end())
							return;
						param = it->second;
						experienceUpgradeRequestMap.erase(it);
					};
					param.timeout.cancel();
					param.onresult(defines::ErrorSource::LIMAX, defines::ErrorCodes::AUANY_EXPERIENCE_UPGRADE_MAMAGER_CLOSE);
				});
			}

			void cleanup()
			{
				hashmap<int64_t, ExperienceUpgradeParam> tempExperienceUpgradeRequestMap;
				{
					std::lock_guard<std::mutex> __lg__(locker);
					experienceUpgradeRequestMap.swap(tempExperienceUpgradeRequestMap);
				}
				for (auto& it : tempExperienceUpgradeRequestMap)
					it.second.timeout.cancel();
			}
		} g_expansion_impl;

		void ExpansionImplCleanup()
		{
			g_expansion_impl.cleanup();
		}
	} // namespace helper {

	Expansion::Expansion()
		: manager(Endpoint::getDefaultEndpointManager())
	{}
	Expansion::~Expansion() {}

	Expansion::Expansion(EndpointManager* _manager)
		: manager(_manager)
	{}

	Expansion::Expansion(const Expansion& src)
		: manager(src.manager)
	{}
	Expansion::Expansion(const Expansion&& src)
		: manager(std::move(src.manager))
	{}
	Expansion& Expansion::operator=(const Expansion& src)
	{
		manager = src.manager;
		return *this;
	}
	Expansion& Expansion::operator=(const Expansion&& src)
	{
		manager = std::move(src.manager);
		return *this;
	}

	void Expansion::experienceUpgrade(const std::string& username, const std::string& token, const std::string& platflag, std::function<void(int, int)> onresult)
	{
		helper::g_expansion_impl.experienceUpgrade(manager, username, token, platflag, onresult);
	}

} // namespace limax {

namespace limax {
	namespace endpoint {
		namespace auanyviews {

			void ExperienceUpgrade::onOpen(const std::vector<int64_t>& sessionids)
			{
				limax::helper::g_expansion_impl.onViewOpen(this);
			}
			void ExperienceUpgrade::onAttach(int64_t sessionid) {}
			void ExperienceUpgrade::onDetach(int64_t sessionid, int reason)
			{
				if (reason >= 0)
				{
					//Application Reason
				}
				else
				{
					//Connection abort Reason
				}
			}
			void ExperienceUpgrade::onClose()
			{
				limax::helper::g_expansion_impl.onViewClose(this);
			}

		}
	}
}