
#include "common.h"

namespace limax {

	MarshalStreamSource::MarshalStreamSource() {}
	MarshalStreamSource::~MarshalStreamSource() {}

	UnmarshalStreamSource::UnmarshalStreamSource() {}
	UnmarshalStreamSource::~UnmarshalStreamSource() {}

} // namespace limax {

namespace limax {

	MarshalStream::MarshalStream(std::shared_ptr<MarshalStreamSource> src)
		: source(src)
	{}

	MarshalStream::MarshalStream(MarshalStreamSource& src)
		: source(std::shared_ptr<MarshalStreamSource>(&src, [](MarshalStreamSource*){}))
	{}

	MarshalStream::MarshalStream(Octets& src)
		: source(std::make_shared<OctetsMarshalStreamSource>(src))
	{}

	MarshalStream::~MarshalStream() {}

	MarshalStream& MarshalStream::push(bool x)
	{
		return push_byte((int8_t)(x ? 1 : 0));
	}

	MarshalStream& MarshalStream::push(int8_t x)
	{
		return push_byte(x);
	}
	MarshalStream& MarshalStream::push(uint8_t x)
	{
		return push_byte(x);
	}
	MarshalStream& MarshalStream::push(int16_t x)
	{
		return push_byte(htobe16(x));
	}
	MarshalStream& MarshalStream::push(uint16_t x)
	{
		return push_byte(htobe16(x));
	}
	MarshalStream& MarshalStream::push(int32_t x)
	{
		return push_byte(htobe32(x));
	}
	MarshalStream& MarshalStream::push(uint32_t x)
	{
		return push_byte(htobe32(x));
	}
	MarshalStream& MarshalStream::push(int64_t x)
	{
		return push_byte(htobe64(x));
	}
	MarshalStream& MarshalStream::push(uint64_t x)
	{
		return push_byte(htobe64(x));
	}

	MarshalStream& MarshalStream::push(float x)
	{
		uint32_t y = aliasing_cast<uint32_t>(x);
		return push(y);
	}
	MarshalStream& MarshalStream::push(double x)
	{
		uint64_t y = aliasing_cast<uint64_t>(x);
		return push(y);
	}

	MarshalStream& MarshalStream::push(const Marshal& x)
	{
		return x.marshal(*this);
	}
	MarshalStream& MarshalStream::push(const Octets& x)
	{
		uint32_t size = (uint32_t)x.size();
		marshal_size(size);
		source->pushbytes(x.begin(), x.size());
		return *this;
	}

	MarshalStream& MarshalStream::marshal_size(int32_t x)
	{
		if (x >= 0)
		{
			if (x < 0x80)
				return push((uint8_t)x);
			if (x < 0x4000)
				return push((uint8_t)((x >> 8) | 0x80)).push((uint8_t)x);
			if (x < 0x200000)
				return push((uint8_t)((x >> 16) | 0xc0)).push((uint8_t)(x >> 8)).push((uint8_t)x);
			if (x < 0x10000000)
				return push((uint8_t)((x >> 24) | 0xe0)).push((uint8_t)(x >> 16)).push((uint8_t)(x >> 8)).push((uint8_t)x);
		}
		return push((uint8_t)0xf0).push(x);
	}
} // namespace limax {

namespace limax {

	UnmarshalStream::UnmarshalStream(std::shared_ptr<UnmarshalStreamSource> src)
		: source(src)
	{}
	UnmarshalStream::UnmarshalStream(UnmarshalStreamSource& src)
		: source(std::shared_ptr<UnmarshalStreamSource>(&src, [](UnmarshalStreamSource*){}))
	{}

	UnmarshalStream::UnmarshalStream(Octets& src)
		: source(std::shared_ptr<UnmarshalStreamSource>(new OctetsUnmarshalStreamSource(src)))
	{}

	UnmarshalStream::~UnmarshalStream() {}

	uint8_t UnmarshalStream::pop_byte_8() const
	{
		uint8_t c;
		pop_byte(c);
		return c;
	}
	uint16_t UnmarshalStream::pop_byte_16() const
	{
		uint16_t s;
		pop_byte(s);
		return be16toh(s);
	}
	uint32_t UnmarshalStream::pop_byte_32() const
	{
		uint32_t l;
		pop_byte(l);
		return be32toh(l);
	}
	uint64_t UnmarshalStream::pop_byte_64() const
	{
		uint64_t ll;
		pop_byte(ll);
		return be64toh(ll);
	}

	bool UnmarshalStream::unmarshal_bool() const
	{
		return !!pop_byte_8();
	}
	int8_t UnmarshalStream::unmarshal_byte() const
	{
		return (int8_t)pop_byte_8();
	}
	uint8_t UnmarshalStream::pop_uint_8() const
	{
		return pop_byte_8();
	}
	int16_t UnmarshalStream::unmarshal_short() const
	{
		return (int16_t)pop_byte_16();
	}
	uint16_t UnmarshalStream::pop_uint_16() const
	{
		return pop_byte_16();
	}
	int32_t UnmarshalStream::unmarshal_int() const
	{
		return (int32_t)pop_byte_32();
	}
	uint32_t UnmarshalStream::pop_uint_32() const
	{
		return pop_byte_32();
	}
	int64_t UnmarshalStream::unmarshal_long() const
	{
		return (int64_t)pop_byte_64();
	}
	uint64_t UnmarshalStream::pop_uint_64() const
	{
		return pop_byte_64();
	}
	int32_t UnmarshalStream::unmarshal_size() const
	{
		int32_t b0 = pop_byte_8();
		if ((b0 & 0x80) == 0)
			return b0;
		if ((b0 & 0x40) == 0) {
			int32_t b1 = pop_byte_8();
			return ((b0 & 0x3f) << 8) | b1;
		}
		if ((b0 & 0x20) == 0) {
			int32_t b1 = pop_byte_8();
			int32_t b2 = pop_byte_8();
			return ((b0 & 0x1f) << 16) | (b1 << 8) | b2;
		}
		if ((b0 & 0x10) == 0) {
			int32_t b1 = pop_byte_8();
			int32_t b2 = pop_byte_8();
			int32_t b3 = pop_byte_8();
			return ((b0 & 0x0f) << 24) | (b1 << 16) | (b2 << 8) | b3;
		}
		return unmarshal_int();
	}
	float UnmarshalStream::unmarshal_float() const
	{
		uint32_t l = pop_byte_32();
		return aliasing_cast<float>(l);
	}
	double UnmarshalStream::unmarshal_double() const
	{
		uint64_t l = pop_byte_64();
		return aliasing_cast<double>(l);
	}
	void UnmarshalStream::unmarshal(const Marshal& x) const
	{
		remove_const(x).unmarshal(*this);
	}
	void UnmarshalStream::unmarshal_Octets(Octets& x) const
	{
		size_t size = (size_t)unmarshal_size();
		const void* data = source->pop(size);
		x.replace(data, size);
	}
	void UnmarshalStream::pop_byte(char* x, size_t size) const
	{
		const void* data = source->pop(size);
		memcpy(x, data, size);
	}

	void UnmarshalStream::transaction(UnmarshalStreamSource::Transaction trans) const
	{
		return source->transaction(trans);
	}

} // namespace limax {

namespace limax {

	Marshal::Marshal() {}
	Marshal::~Marshal() {}

	std::ostream& Marshal::trace(std::ostream& os) const
	{
		return os;
	}

} // namespace limax {

namespace limax {

	OctetsMarshalStreamSource::OctetsMarshalStreamSource(Octets& src)
		: data(src)
	{}
	OctetsMarshalStreamSource::~OctetsMarshalStreamSource() {}

	void OctetsMarshalStreamSource::pushbytes(const void* _data, size_t _size)
	{
		data.insert(data.end(), _data, _size);
	}

} // namespace limax {

namespace limax {

	OctetsUnmarshalStreamSource::OctetsUnmarshalStreamSource(const Octets& src)
		: pos(0), tranpos(0), data(src)
	{}
	OctetsUnmarshalStreamSource::~OctetsUnmarshalStreamSource() {}

	const void* OctetsUnmarshalStreamSource::pop(size_t size) const
	{
		if ((pos + size) > data.size())
			throw MarshalException();
		const void* _data = (const char*)data.begin() + pos;
		pos += size;
		return _data;
	}

	void OctetsUnmarshalStreamSource::transaction(Transaction trans) const
	{
		switch (trans)
		{
		case Begin:
			tranpos = pos;
			break;
		case Rollback:
			pos = tranpos;
			break;
		case Commit:
			if (pos >= MAXSPARE)
			{
				Octets& data = remove_const(this->data);
				data.erase(data.begin(), (char*)data.begin() + pos);
				pos = 0;
			}
		default:
			break;
		}
	}

	bool OctetsUnmarshalStreamSource::eos() const
	{
		return data.size() == pos;
	}

	size_t OctetsUnmarshalStreamSource::position() const
	{
		return pos;
	}

	OctetsMarshalStream::OctetsMarshalStream()
		: MarshalStream(source), source(data)
	{}

	OctetsMarshalStream::~OctetsMarshalStream() {}

	Octets OctetsMarshalStream::getOctets() const
	{
		return data;
	}

} // namespace limax {

