﻿
using limax.util;
using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace limax.codec
{
    public abstract class Codec : IDisposable
    {
        public abstract void update(byte c);
        public abstract void update(byte[] data, int off, int len);
        public abstract void flush();
        public virtual void Dispose() { }
        private class _nullCodec : Codec
        {
            public override void update(byte c) { }
            public override void update(byte[] data, int off, int len) { }
            public override void flush() { }
        }
        private static readonly _nullCodec _null = new _nullCodec();
        public static Codec Null() { return _null; }
    }
    public abstract class MD : Codec
    {
        private readonly Codec sink;
        private byte[] oneByte = new byte[1];
        private readonly HashAlgorithm md;
        public MD(Codec sink, HashAlgorithm md)
        {
            this.sink = sink;
            this.md = md;
            md.Initialize();
        }
        public override void update(byte c)
        {
            oneByte[0] = c;
            md.TransformBlock(oneByte, 0, 1, oneByte, 0);
            sink.update(c);
        }
        public override void update(byte[] data, int off, int len)
        {
            md.TransformBlock(data, off, len, data, off);
            sink.update(data, off, len);
        }
        public override void flush()
        {
            sink.flush();
        }
        public byte[] digest()
        {
            md.TransformFinalBlock(new byte[0], 0, 0);
            return md.Hash;
        }
        public override void Dispose()
        {
            md.Dispose();
            sink.Dispose();
        }
    }
    public sealed class MD5 : MD
    {
        public MD5(Codec sink) : base(sink, System.Security.Cryptography.MD5.Create()) { }
        public static byte[] digest(byte[] message)
        {
            using (var md = System.Security.Cryptography.MD5.Create())
                return md.ComputeHash(message);
        }
    }
    public sealed class SHA1 : MD
    {
        public SHA1(Codec sink) : base(sink, System.Security.Cryptography.SHA1.Create()) { }
        public static byte[] digest(byte[] message)
        {
            using (var md = System.Security.Cryptography.SHA1.Create())
                return md.ComputeHash(message);
        }
    }
    public sealed class HmacMD5 : MD
    {
        public HmacMD5(Codec sink, byte[] key, int off, int len) : base(sink, new HMACMD5(Helper.copyOfRange(key, off, off + len))) { }
    }
    public sealed class HmacSHA1 : MD
    {
        public HmacSHA1(Codec sink, byte[] key, int off, int len) : base(sink, new HMACSHA1(Helper.copyOfRange(key, off, off + len))) { }
    }
    public sealed class Encrypt : Codec
    {
        private readonly Codec sink;
        private readonly ICryptoTransform cipher;
        private readonly byte[] _iv;
        private readonly byte[] _in = new byte[16];
        private readonly byte[] _out = new byte[16];
        private int count = 0;
        public Encrypt(Codec sink, byte[] key)
        {
            this.sink = sink;
            _iv = MD5.digest(key);
            AesManaged aes = new AesManaged();
            aes.Mode = CipherMode.ECB;
            cipher = aes.CreateEncryptor(_iv, _iv);
        }
        private void succeed()
        {
            cipher.TransformBlock(_iv, 0, 16, _iv, 0);
        }
        public override void update(byte c)
        {
            if (count < 0)
            {
                sink.update(_iv[count++ + 16] ^= c);
                return;
            }
            _in[count++] = c;
            if (count < 16)
                return;
            succeed();
            for (int i = 0; i < 16; i++)
                _iv[i] ^= _in[i];
            sink.update(_iv, 0, 16);
            count = 0;
        }
        public override void update(byte[] data, int off, int len)
        {
            int i = off;
            len += off;
            if (count < 0)
            {
                for (; i < len && count < 0; i++, count++)
                    sink.update(_iv[count + 16] ^= data[i]);
            }
            else if (count > 0)
            {
                for (; i < len && count < 16; i++, count++)
                    _in[count] = data[i];
                if (count < 16)
                    return;
                succeed();
                for (int j = 0; j < 16; j++)
                    _iv[j] ^= _in[j];
                sink.update(_iv, 0, 16);
                count = 0;
            }
            int nblocks = (len - i) >> 4;
            for (int j = 0; j < nblocks; j++)
            {
                succeed();
                for (int k = 0; k < 16; k++)
                    _iv[k] ^= data[i + j * 16 + k];
                sink.update(_iv, 0, 16);
            }
            for (i += nblocks << 4; i < len; i++)
                _in[count++] = data[i];
        }
        public override void flush()
        {
            if (count > 0)
            {
                succeed();
                for (int i = 0; i < count; i++)
                    sink.update(_iv[i] ^= _in[i]);
                count -= 16;
            }
            sink.flush();
        }
        public override void Dispose()
        {
            cipher.Dispose();
            sink.Dispose();
        }
    }
    public sealed class Decrypt : Codec
    {
        private readonly Codec sink;
        private readonly ICryptoTransform cipher;
        private readonly byte[] _iv;
        private readonly byte[] _in = new byte[16];
        private readonly byte[] _out = new byte[16];
        private int count = 0;
        public Decrypt(Codec sink, byte[] key)
        {
            this.sink = sink;
            _iv = MD5.digest(key);
            AesManaged aes = new AesManaged();
            aes.Mode = CipherMode.ECB;
            cipher = aes.CreateEncryptor(_iv, _iv);
        }
        private void succeed()
        {
            cipher.TransformBlock(_iv, 0, 16, _iv, 0);
        }
        public override void update(byte c)
        {
            if (count < 0)
            {
                sink.update((byte)(_iv[count + 16] ^ c));
                _iv[count++ + 16] = c;
                return;
            }
            _in[count++] = c;
            if (count < 16)
                return;
            succeed();
            for (int i = 0; i < 16; i++)
            {
                _out[i] = (byte)(_iv[i] ^ _in[i]);
                _iv[i] = _in[i];
            }
            sink.update(_out, 0, 16);
            count = 0;
        }
        public override void update(byte[] data, int off, int len)
        {
            int i = off;
            len += off;
            if (count < 0)
            {
                for (; i < len && count < 0; i++, count++)
                {
                    sink.update((byte)(_iv[count + 16] ^ data[i]));
                    _iv[count + 16] = data[i];
                }
            }
            else if (count > 0)
            {
                for (; i < len && count < 16; i++, count++)
                    _in[count] = data[i];
                if (count < 16)
                    return;
                succeed();
                for (int j = 0; j < 16; j++)
                {
                    _out[j] = (byte)(_iv[j] ^ _in[j]);
                    _iv[j] = _in[j];
                }
                sink.update(_out, 0, 16);
                count = 0;
            }
            int nblocks = (len - i) >> 4;
            for (int j = 0; j < nblocks; j++)
            {
                succeed();
                for (int k = 0; k < 16; k++)
                {
                    byte c = data[i + j * 16 + k];
                    _out[k] = (byte)(_iv[k] ^ c);
                    _iv[k] = c;
                }
                sink.update(_out, 0, 16);
            }
            for (i += nblocks << 4; i < len; i++)
                _in[count++] = data[i];
        }
        public override void flush()
        {
            if (count > 0)
            {
                succeed();
                for (int i = 0; i < count; i++)
                {
                    sink.update((byte)(_iv[i] ^ _in[i]));
                    _iv[i] = _in[i];
                }
                count -= 16;
            }
            sink.flush();
        }
        public override void Dispose()
        {
            cipher.Dispose();
            sink.Dispose();
        }
    }
    public sealed class BufferedSink : Codec
    {
        private readonly Codec sink;
        private readonly byte[] buffer = new byte[8192];
        private int pos = 0;
        private void flushInternal()
        {
            if (pos > 0)
            {
                sink.update(buffer, 0, pos);
                pos = 0;
            }
        }
        public BufferedSink(Codec sink)
        {
            this.sink = sink;
        }
        public override void update(byte c)
        {
            if (buffer.Length == pos)
            {
                flushInternal();
            }
            buffer[pos++] = c;
        }
        public override void update(byte[] data, int off, int len)
        {
            if (len >= buffer.Length)
            {
                flushInternal();
                sink.update(data, off, len);
                return;
            }
            if (len > buffer.Length - pos)
            {
                flushInternal();
            }
            Buffer.BlockCopy(data, off, buffer, pos, len);
            pos += len;
        }
        public override void flush()
        {
            flushInternal();
            sink.flush();
        }
        public override void Dispose()
        {
            sink.Dispose();
        }
    }
    public sealed class RFC2118Encode : Codec
    {
        private readonly Codec sink;
        private int pos = 0;
        private int rem = 0;
        private byte[] dict = new byte[8192];
        private short[] hash = new short[65536];
        private int idx = 0;
        private int match_idx;
        private int match_off = -1;
        private int match_len;
        private bool flushed = true;
        public RFC2118Encode(Codec sink)
        {
            this.sink = sink;
            for (int i = 0; i < hash.Length; i++)
                hash[i] = -1;
        }
        private void putBits(int val, int nbits)
        {
            pos += nbits;
            rem |= val << (32 - pos);
            while (pos > 7)
            {
                sink.update((byte)(rem >> 24));
                pos -= 8;
                rem <<= 8;
            }
        }
        private void putLiteral(byte c)
        {
            if ((c & 0x80) == 0)
                putBits(c, 8);
            else
                putBits(c & 0x7f | 0x100, 9);
        }
        private void putTuple(int off, int len)
        {
            if (off < 64)
                putBits(0x3c0 | off, 10);
            else if (off < 320)
                putBits(0xe00 | (off - 64), 12);
            else
                putBits(0xc000 | (off - 320), 16);
            if (len < 4)
                putBits(0, 1);
            else if (len < 8)
                putBits(0x08 | (len & 0x03), 4);
            else if (len < 16)
                putBits(0x30 | (len & 0x07), 6);
            else if (len < 32)
                putBits(0xe0 | (len & 0x0f), 8);
            else if (len < 64)
                putBits(0x3c0 | (len & 0x1f), 10);
            else if (len < 128)
                putBits(0xf80 | (len & 0x3f), 12);
            else if (len < 256)
                putBits(0x3f00 | (len & 0x7f), 14);
            else if (len < 512)
                putBits(0xfe00 | (len & 0xff), 16);
            else if (len < 1024)
                putBits(0x3fc00 | (len & 0x1ff), 18);
            else if (len < 2048)
                putBits(0xff800 | (len & 0x3ff), 20);
            else if (len < 4096)
                putBits(0x3ff000 | (len & 0x7ff), 22);
            else if (len < 8192)
                putBits(0xffe000 | (len & 0xfff), 24);
        }
        private void _flush()
        {
            if (match_off > 0)
            {
                if (match_len == 2)
                {
                    putLiteral(dict[match_idx - 2]);
                    putLiteral(dict[match_idx - 1]);
                }
                else
                    putTuple(match_off, match_len);
                match_off = -1;
            }
            else
                putLiteral(dict[idx - 1]);
            flushed = true;
        }
        public override void update(byte c)
        {
            if (idx == dict.Length)
            {
                if (!flushed)
                    _flush();
                for (int i = 0; i < hash.Length; i++)
                    hash[i] = -1;
                idx = 0;
            }
            dict[idx++] = c;
            if (flushed)
            {
                flushed = false;
                return;
            }
            int key = ((c & 0xff) << 8) | dict[idx - 2] & 0xff;
            int tmp = hash[key];
            hash[key] = (short)idx;
            if (match_off > 0)
            {
                if (dict[match_idx] == c)
                {
                    match_idx++;
                    match_len++;
                }
                else
                {
                    if (match_len == 2)
                    {
                        putLiteral(dict[match_idx - 2]);
                        putLiteral(dict[match_idx - 1]);
                    }
                    else
                        putTuple(match_off, match_len);
                    match_off = -1;
                }
            }
            else
            {
                if (tmp != -1)
                {
                    match_idx = tmp;
                    match_off = idx - tmp;
                    match_len = 2;
                }
                else
                    putLiteral(dict[idx - 2]);
            }
        }
        public override void update(byte[] data, int off, int len)
        {
            for (int i = off; i < len; i++)
                update(data[i]);
        }
        public override void flush()
        {
            if (!flushed)
            {
                _flush();
                if (pos > 0)
                    putBits(0x3c0, 10);
            }
            sink.flush();
        }
        public override void Dispose()
        {
            sink.Dispose();
        }
    }
    public sealed class RFC2118Decode : Codec
    {
        private readonly Codec sink;
        private int rem = 0;
        private int pos = 0;
        private int off = -1;
        private int len;
        private byte[] hist = new byte[8192 * 3];
        private int hpos = 0;
        public sealed class UncompressException : Exception { }
        public RFC2118Decode(Codec sink)
        {
            this.sink = sink;
        }
        private void drain()
        {
            if (hpos >= 8192 * 2)
            {
                Buffer.BlockCopy(hist, hpos - 8192, hist, 0, 8192);
                hpos = 8192;
            }
        }
        private void copy(int dstPos, int srcPos, int length)
        {
            for (int i = 0; i < length; i++)
                hist[dstPos++] = hist[srcPos++];
        }
        private void output(byte c)
        {
            sink.update(hist[hpos++] = c);
            drain();
        }
        private void output(int off, int len)
        {
            if (hpos < off)
                throw new UncompressException();
            copy(hpos, hpos - off, len);
            sink.update(hist, hpos, len);
            hpos += len;
            drain();
        }
        private int bitCompute()
        {
            long val = (rem << (32 - pos)) & 0xffffffffL;
            if (off < 0)
            {
                if (val < 0x80000000L)
                    return 8;
                else if (val < 0xc0000000L)
                    return 9;
                else if (val < 0xe0000000L)
                    return 16;
                else if (val < 0xf0000000L)
                    return 12;
                else
                    return 10;
            }
            else
            {
                if (val < 0x80000000L)
                    return 1;
                else if (val < 0xc0000000L)
                    return 4;
                else if (val < 0xe0000000L)
                    return 6;
                else if (val < 0xf0000000L)
                    return 8;
                else if (val < 0xf8000000L)
                    return 10;
                else if (val < 0xfc000000L)
                    return 12;
                else if (val < 0xfe000000L)
                    return 14;
                else if (val < 0xff000000L)
                    return 16;
                else if (val < 0xff800000L)
                    return 18;
                else if (val < 0xffc00000L)
                    return 20;
                else if (val < 0xffe00000L)
                    return 22;
                else if (val < 0xfff00000L)
                    return 24;
                else
                    return 32;
            }
        }
        private void process()
        {
            long val = (rem << (32 - pos)) & 0xffffffffL;
            if (off < 0)
            {
                if (val < 0x80000000L)
                {
                    output((byte)(val >> 24));
                    pos -= 8;
                }
                else if (val < 0xc0000000L)
                {
                    output((byte)((val >> 23) | 0x80));
                    pos -= 9;
                }
                else if (val < 0xe0000000L)
                {
                    off = (int)(((val >> 16) & 0x1fff) + 320);
                    pos -= 16;
                }
                else if (val < 0xf0000000L)
                {
                    off = (int)(((val >> 20) & 0xff) + 64);
                    pos -= 12;
                }
                else
                {
                    off = (int)((val >> 22) & 0x3f);
                    pos -= 10;
                    if (off == 0)
                        off = -1;
                }
            }
            else
            {
                if (val < 0x80000000L)
                {
                    len = 3;
                    pos -= 1;
                }
                else if (val < 0xc0000000L)
                {
                    len = (int)(4 | ((val >> 28) & 3));
                    pos -= 4;
                }
                else if (val < 0xe0000000L)
                {
                    len = (int)(8 | ((val >> 26) & 7));
                    pos -= 6;
                }
                else if (val < 0xf0000000L)
                {
                    len = (int)(16 | ((val >> 24) & 15));
                    pos -= 8;
                }
                else if (val < 0xf8000000L)
                {
                    len = (int)(32 | ((val >> 22) & 31));
                    pos -= 10;
                }
                else if (val < 0xfc000000L)
                {
                    len = (int)(64 | ((val >> 20) & 63));
                    pos -= 12;
                }
                else if (val < 0xfe000000L)
                {
                    len = (int)(128 | ((val >> 18) & 127));
                    pos -= 14;
                }
                else if (val < 0xff000000L)
                {
                    len = (int)(256 | ((val >> 16) & 255));
                    pos -= 16;
                }
                else if (val < 0xff800000L)
                {
                    len = (int)(512 | ((val >> 14) & 511));
                    pos -= 18;
                }
                else if (val < 0xffc00000L)
                {
                    len = (int)(1024 | ((val >> 12) & 1023));
                    pos -= 20;
                }
                else if (val < 0xffe00000L)
                {
                    len = (int)(2048 | ((val >> 10) & 2047));
                    pos -= 22;
                }
                else if (val < 0xfff00000L)
                {
                    len = (int)(4096 | ((val >> 8) & 4095));
                    pos -= 24;
                }
                else
                    throw new UncompressException();
                output(off, len);
                off = -1;
            }
        }
        public override void update(byte c)
        {
            pos += 8;
            rem = (rem << 8) | (c & 0xff);
            while (pos > 24)
                process();
        }
        public override void update(byte[] data, int off, int len)
        {
            for (int i = off; i < len; i++)
                update(data[i]);
        }
        public override void flush()
        {
            while (pos >= bitCompute())
                process();
            sink.flush();
        }
        public override void Dispose()
        {
            sink.Dispose();
        }
    }
    public sealed class SinkStream : Codec
    {
        private readonly Stream stream;
        public SinkStream(Stream stream)
        {
            this.stream = stream;
        }
        public override void update(byte c)
        {
            stream.WriteByte(c);
        }
        public override void update(byte[] data, int off, int len)
        {
            stream.Write(data, off, len);
        }
        public override void flush()
        {
            stream.Flush();
        }
        public override void Dispose()
        {
            stream.Dispose();
        }
    }
    public sealed class MarshalException : Exception { }

    public sealed class CodecException : Exception
    {
        public CodecException(Exception e) : base("", e) { }
        public CodecException(string message) : base(message) { }
        public CodecException(string message, Exception e) : base(message, e) { }
    }
    public interface Marshal
    {
        OctetsStream marshal(OctetsStream os);
        OctetsStream unmarshal(OctetsStream os);
    }
    public class Octets : IComparable<Octets>
    {
        private const int DEFAULT_SIZE = 128;
        internal protected byte[] buffer = null;
        private int count = 0;
        private byte[] roundup(int size)
        {
            int capacity = DEFAULT_SIZE;
            while (size > capacity)
                capacity <<= 1;
            return new byte[capacity];
        }
        public void reserve(int size)
        {
            if (buffer == null)
            {
                buffer = roundup(size);
            }
            else if (size > buffer.Length)
            {
                byte[] tmp = roundup(size);
                Buffer.BlockCopy(buffer, 0, tmp, 0, count);
                buffer = tmp;
            }
        }
        public Octets replace(byte[] data, int pos, int size)
        {
            reserve(size);
            Buffer.BlockCopy(data, pos, buffer, 0, size);
            count = size;
            return this;
        }
        public Octets replace(Octets data, int pos, int size)
        {
            return replace(data.buffer, pos, size);
        }
        public Octets replace(byte[] data)
        {
            return replace(data, 0, data.Length);
        }
        public Octets replace(Octets data)
        {
            return replace(data.buffer, 0, data.count);
        }
        public Octets()
        {
            reserve(DEFAULT_SIZE);
        }
        public Octets(int size)
        {
            reserve(size);
        }
        public Octets(Octets rhs)
        {
            replace(rhs);
        }
        public Octets(byte[] rhs)
        {
            replace(rhs);
        }
        internal Octets(byte[] bytes, int length)
        {
            this.buffer = bytes;
            this.count = length;
        }
        public static Octets wrap(byte[] bytes, int length)
        {
            return new Octets(bytes, length);
        }
        public static Octets wrap(byte[] bytes)
        {
            return wrap(bytes, bytes.Length);
        }
        public Octets(byte[] rhs, int pos, int size)
        {
            replace(rhs, pos, size);
        }
        public Octets(Octets rhs, int pos, int size)
        {
            replace(rhs, pos, size);
        }
        public Octets resize(int size)
        {
            reserve(size);
            count = size;
            return this;
        }
        public int size()
        {
            return count;
        }
        public int capacity()
        {
            return buffer.Length;
        }
        public Octets clear()
        {
            count = 0;
            return this;
        }
        public Octets swap(Octets rhs)
        {
            int size = count;
            count = rhs.count;
            rhs.count = size;
            byte[] tmp = rhs.buffer;
            rhs.buffer = buffer;
            buffer = tmp;
            return this;
        }
        public Octets push_byte(byte data)
        {
            reserve(count + 1);
            buffer[count++] = data;
            return this;
        }
        public Octets erase(int from, int to)
        {
            Buffer.BlockCopy(buffer, to, buffer, from, count - to);
            count -= to - from;
            return this;
        }
        public Octets insert(int from, byte[] data, int pos, int size)
        {
            reserve(count + size);
            Buffer.BlockCopy(buffer, from, buffer, from + size, count - from);
            Buffer.BlockCopy(data, pos, buffer, from, size);
            count += size;
            return this;
        }
        public Octets insert(int from, Octets data, int pos, int size)
        {
            return insert(from, data.buffer, pos, size);
        }
        public Octets insert(int from, byte[] data)
        {
            return insert(from, data, 0, data.Length);
        }
        public Octets insert(int from, Octets data)
        {
            return insert(from, data.buffer, 0, data.size());
        }
        public byte[] getBytes()
        {
            byte[] tmp = new byte[count];
            Buffer.BlockCopy(buffer, 0, tmp, 0, count);
            return tmp;
        }
        public byte[] array()
        {
            return buffer;
        }
        public byte getByte(int pos)
        {
            return buffer[pos];
        }
        public void setByte(int pos, byte b)
        {
            buffer[pos] = b;
        }
        public int CompareTo(Octets rhs)
        {
            int c = count - rhs.count;
            if (c != 0)
                return c;
            byte[] v1 = buffer;
            byte[] v2 = rhs.buffer;
            for (int i = 0; i < count; i++)
            {
                if (v1[i] > v2[i])
                    return 1;
                else if (v1[i] < v2[i])
                    return -1;
            }
            return 0;
        }
        public override bool Equals(object obj)
        {
            if (object.ReferenceEquals(this, obj))
                return true;
            Octets o = obj as Octets;
            return o != null ? 0 == CompareTo(o) : false;
        }
        public override int GetHashCode()
        {
            int result = 1;
            for (int i = 0; i < count; i++)
                result = 31 * result + buffer[i];
            return result;
        }
    }
    public sealed class OctetsStream : Octets
    {
        private const int MAXSPARE = 8192;
        private int pos = 0;
        private int tranpos = 0;
        public OctetsStream()
        {
        }
        public OctetsStream(int size)
            : base(size)
        {
        }
        public OctetsStream(Octets o)
            : base(o)
        {
        }
        private OctetsStream(byte[] bytes, int length)
            : base(bytes, length)
        {
        }
        public static OctetsStream wrap(Octets o)
        {
            return new OctetsStream(o.array(), o.size());
        }
        public bool eos()
        {
            return pos == size();
        }
        public OctetsStream position(int pos)
        {
            this.pos = pos;
            return this;
        }
        public int position()
        {
            return pos;
        }
        public int remain()
        {
            return size() - pos;
        }
        public OctetsStream begin()
        {
            tranpos = pos;
            return this;
        }
        public OctetsStream rollback()
        {
            pos = tranpos;
            return this;
        }
        public OctetsStream commit()
        {
            if (pos >= MAXSPARE)
            {
                erase(0, pos);
                pos = 0;
            }
            return this;
        }
        private OctetsStream push_bytes(byte[] data)
        {
            insert(size(), data);
            return this;
        }
        public OctetsStream marshal(byte x)
        {
            push_byte(x);
            return this;
        }
        public OctetsStream marshal(bool b)
        {
            push_byte((byte)(b ? 1 : 0));
            return this;
        }
        public OctetsStream marshal(short x)
        {
            return marshal((byte)(x >> 8)).marshal((byte)(x));
        }
        public OctetsStream marshal(char x)
        {
            return marshal((byte)(x >> 8)).marshal((byte)(x));
        }
        public OctetsStream marshal(int x)
        {
            return marshal((byte)(x >> 24)).marshal((byte)(x >> 16)).marshal((byte)(x >> 8)).marshal((byte)(x));
        }
        public OctetsStream marshal(long x)
        {
            return marshal((byte)(x >> 56)).marshal((byte)(x >> 48)).marshal((byte)(x >> 40)).marshal((byte)(x >> 32)).marshal((byte)(x >> 24)).marshal((byte)(x >> 16)).marshal((byte)(x >> 8)).marshal((byte)(x));
        }
        public OctetsStream marshal(float x)
        {
            byte[] bytes = BitConverter.GetBytes(x);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(bytes);
            return push_bytes(bytes);
        }
        public OctetsStream marshal(double x)
        {
            byte[] bytes = BitConverter.GetBytes(x);
            if (BitConverter.IsLittleEndian)
                Array.Reverse(bytes);
            return push_bytes(bytes);
        }
        public OctetsStream marshal(Marshal m)
        {
            return m.marshal(this);
        }
        public OctetsStream marshal_size(int x)
        {
            if (x >= 0)
            {
                if (x < 0x80) // 0xxxxxxx
                    return marshal((byte)x);
                if (x < 0x4000) // 10xxxxxx xxxxxxxx
                    return marshal((byte)((x >> 8) | 0x80)).marshal((byte)x);
                if (x < 0x200000) // 110xxxxx xxxxxxxx xxxxxxxx
                    return marshal((byte)((x >> 16) | 0xc0)).marshal(
                            (byte)(x >> 8)).marshal((byte)x);
                if (x < 0x10000000) // 1110xxxx xxxxxxxx xxxxxxxx xxxxxxxx
                    return marshal((byte)((x >> 24) | 0xe0))
                            .marshal((byte)(x >> 16)).marshal((byte)(x >> 8))
                            .marshal((byte)x);
            }
            return marshal((byte)0xf0).marshal(x);
        }
        public OctetsStream marshal(Octets o)
        {
            this.marshal_size(o.size());
            insert(size(), o);
            return this;
        }
        public OctetsStream marshal(byte[] bytes)
        {
            this.marshal_size(bytes.Length);
            insert(size(), bytes);
            return this;
        }
        public OctetsStream marshal(string str)
        {
            return marshal(Encoding.UTF8.GetBytes(str));
        }
        public byte unmarshal_byte()
        {
            if (pos + 1 > size())
                throw new MarshalException();
            return getByte(pos++);
        }
        public bool unmarshal_bool()
        {
            return unmarshal_byte() == 1;
        }
        public short unmarshal_short()
        {
            if (pos + 2 > size())
                throw new MarshalException();
            byte b0 = getByte(pos++);
            byte b1 = getByte(pos++);
            return (short)((b0 << 8) | (b1 & 0xff));
        }
        public char unmarshal_char()
        {
            if (pos + 2 > size())
                throw new MarshalException();
            byte b0 = getByte(pos++);
            byte b1 = getByte(pos++);
            return (char)((b0 << 8) | (b1 & 0xff));
        }
        public int unmarshal_int()
        {
            if (pos + 4 > size())
                throw new MarshalException();
            byte b0 = getByte(pos++);
            byte b1 = getByte(pos++);
            byte b2 = getByte(pos++);
            byte b3 = getByte(pos++);
            return ((b0 & 0xff) << 24) | ((b1 & 0xff) << 16) | ((b2 & 0xff) << 8) | (b3 & 0xff);
        }
        public long unmarshal_long()
        {
            if (pos + 8 > size())
                throw new MarshalException();
            byte b0 = getByte(pos++);
            byte b1 = getByte(pos++);
            byte b2 = getByte(pos++);
            byte b3 = getByte(pos++);
            byte b4 = getByte(pos++);
            byte b5 = getByte(pos++);
            byte b6 = getByte(pos++);
            byte b7 = getByte(pos++);
            return (((long)b0 & 0xff) << 56) | (((long)b1 & 0xff) << 48) | (((long)b2 & 0xff) << 40) | (((long)b3 & 0xff) << 32) | (((long)b4 & 0xff) << 24) | (((long)b5 & 0xff) << 16) | (((long)b6 & 0xff) << 8) | ((long)b7 & 0xff);
        }
        public float unmarshal_float()
        {
            if (pos + 4 > size())
                throw new MarshalException();
            float v;
            if (BitConverter.IsLittleEndian)
            {
                byte[] data = new byte[] { buffer[pos + 3], buffer[pos + 2], buffer[pos + 1], buffer[0] };
                v = BitConverter.ToSingle(data, 0);
            }
            else
            {
                v = BitConverter.ToSingle(buffer, pos);
            }
            pos += 4;
            return v;
        }
        public double unmarshal_double()
        {
            if (pos + 8 > size())
                throw new MarshalException();
            double v;
            if (BitConverter.IsLittleEndian)
            {
                byte[] data = new byte[] { buffer[pos + 7], buffer[pos + 6], buffer[pos + 5], buffer[pos + 4], buffer[pos + 3], buffer[pos + 2], buffer[pos + 1], buffer[0] };
                v = BitConverter.ToSingle(data, 0);
            }
            else
            {
                v = BitConverter.ToSingle(buffer, pos);
            }
            pos += 8;
            return v;
        }
        public int unmarshal_size()
        {
            byte b0 = unmarshal_byte();
            if ((b0 & 0x80) == 0)
                return b0;
            if ((b0 & 0x40) == 0)
            {
                byte b1 = unmarshal_byte();
                return ((b0 & 0x3f) << 8) | (b1 & 0xff);
            }
            if ((b0 & 0x20) == 0)
            {
                byte b1 = unmarshal_byte();
                byte b2 = unmarshal_byte();
                return ((b0 & 0x1f) << 16) | ((b1 & 0xff) << 8) | (b2 & 0xff);
            }
            if ((b0 & 0x10) == 0)
            {
                byte b1 = unmarshal_byte();
                byte b2 = unmarshal_byte();
                byte b3 = unmarshal_byte();
                return ((b0 & 0x0f) << 24) | ((b1 & 0xff) << 16)
                        | ((b2 & 0xff) << 8) | (b3 & 0xff);
            }
            return unmarshal_int();
        }
        public Octets unmarshal_Octets()
        {
            int _size = this.unmarshal_size();
            if (pos + _size > size())
                throw new MarshalException();
            Octets o = new Octets(this, pos, _size);
            pos += _size;
            return o;
        }
        public byte[] unmarshal_bytes()
        {
            int _size = this.unmarshal_size();
            if (pos + _size > size())
                throw new MarshalException();
            byte[] copy = new byte[_size];
            Buffer.BlockCopy(buffer, pos, copy, 0, _size);
            pos += _size;
            return copy;
        }
        public string unmarshal_string()
        {
            int _size = this.unmarshal_size();
            if (pos + _size > size())
                throw new MarshalException();
            string v = Encoding.UTF8.GetString(array(), pos, _size);
            pos += _size;
            return v;
        }
        public OctetsStream unmarshal(Marshal m)
        {
            return m.unmarshal(this);
        }
    }
}