package limax.edb;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;

final class ChainPage implements Page.Impl {
	@SuppressWarnings("rawtypes")
	final private static class ChainHDR implements Comparable {
		private PageCache cache;
		private long dataAddr = 0;
		private byte prefixLen = 0;
		private byte prefixKey[] = new byte[15];

		void decode(ByteBuffer bb) {
			dataAddr = bb.getLong();
			prefixLen = bb.get();
			bb.get(prefixKey);
		}

		void encode(ByteBuffer bb) {
			bb.putLong(dataAddr);
			bb.put(prefixLen);
			bb.put(prefixKey);
		}

		@Override
		public int compareTo(Object o) {
			byte key[] = (byte[]) o;
			if (prefixKey.length != prefixLen) { // prefix sufficient
				int c = IndexPage.compare(prefixKey, key,
						Math.min(prefixLen, key.length));
				return c != 0 ? c : prefixLen - key.length;
			}
			if (prefixKey.length >= key.length) {
				int c = IndexPage.compare(prefixKey, key, key.length);
				return c != 0 ? c : prefixKey.length - key.length;
			}
			int c = IndexPage.compare(prefixKey, key, prefixKey.length);
			if (c != 0)
				return c;
			try {
				byte ekey[] = DataPage.getKey(cache, dataAddr);
				c = IndexPage.compare(ekey, key,
						Math.min(ekey.length, key.length));
				return c != 0 ? c : ekey.length - key.length;
			} catch (Exception e) {
			}
			return 0;
		}
	}

	final private static int ROW_LENGTH = Configure.CHAIN_ROW_LENGTH;
	final private static int posPrevIndex = Configure.PAGESIZE
			- Configure.CHAIN_TAIL_LENGTH;
	final private static int CHAINCOUNT = posPrevIndex / ROW_LENGTH;
	final private static int posNextIndex = posPrevIndex + 8;
	final private static int posLPos = posNextIndex + 8;
	final private static int posRPos = posLPos + 2;
	final private static long primes[] = { 17, 53, 97, 193, 389, 769, 1543,
			3079, 6151, 12289 };
	final private static Queue<ChainPage> pool = new ConcurrentLinkedQueue<>();
	final private ReentrantLock headerLock = new ReentrantLock();
	private PageCache cache;
	private Page page;
	private ChainHDR header[] = new ChainHDR[CHAINCOUNT];
	private long prevIndex;
	private long nextIndex;
	private short l_pos;
	private short r_pos;

	Page getPage() {
		return page;
	}

	private ChainPage() {
		for (int i = 0; i < header.length; i++)
			header[i] = new ChainHDR();
	}

	private static ChainPage _alloc() {
		ChainPage obj = pool.poll();
		if (obj == null)
			obj = new ChainPage();
		return obj;
	}

	static ChainPage alloc(PageCache cache, Page page) {
		ChainPage obj = _alloc();
		obj.init(cache, page);
		return obj;
	}

	static ChainPage alloc(PageCache cache, Page page, PageLayout layout) {
		ChainPage obj = _alloc();
		obj.init(cache, page, layout);
		return obj;
	}

	@Override
	public void free() {
		pool.offer(this);
	}

	@Override
	public PageLayout createSnapshot() {
		PageLayout layout = new PageLayout();
		ByteBuffer bb = layout.data();
		bb.position(l_pos * ROW_LENGTH);
		for (int i = l_pos; i < r_pos; i++)
			header[i].encode(bb);
		bb.putLong(posPrevIndex, prevIndex);
		bb.putLong(posNextIndex, nextIndex);
		bb.putShort(posLPos, l_pos);
		bb.putShort(posRPos, r_pos);
		return layout;
	}

	private void init(PageCache cache, Page page) {
		this.cache = cache;
		this.page = page;
		this.prevIndex = 0;
		this.nextIndex = 0;
		this.l_pos = 0;
		this.r_pos = 0;
		for (int i = 0; i < header.length; i++)
			header[i].cache = cache;
	}

	private void init(PageCache cache, Page page, PageLayout layout) {
		this.cache = cache;
		this.page = page;
		for (int i = 0; i < header.length; i++)
			header[i].cache = cache;
		ByteBuffer bb = layout.data();
		prevIndex = bb.getLong(posPrevIndex);
		nextIndex = bb.getLong(posNextIndex);
		l_pos = bb.getShort(posLPos);
		r_pos = bb.getShort(posRPos);
		bb.position(l_pos * ROW_LENGTH);
		for (int i = l_pos; i < r_pos; i++)
			header[i].decode(bb);
		layout.free();
	}

	private int chainSize() {
		return r_pos - l_pos;
	}

	private static ChainPage RC_Page(PageCache cache, long index)
			throws IOException {
		return cache.rLoad(index).asChainPage();
	}

	private ChainPage RC_Page(long index) throws IOException {
		return cache.rLoad(index).asChainPage();
	}

	private ChainPage WC_Page(long index) throws IOException {
		return cache.wLoad(index).asChainPage();
	}

	Cursor find(final byte[] key, Cursor r) {
		int pos = Arrays.binarySearch(header, l_pos, r_pos, key);
		if (pos >= 0)
			return r.set(pos, header[pos].dataAddr);
		return r.set(-pos - 1, 0);
	}

	byte[] getKey(int pos) throws IOException {
		int prefixLen = header[pos].prefixLen;
		byte[] prefixKey = header[pos].prefixKey;
		return prefixLen < prefixKey.length ? Arrays.copyOf(prefixKey,
				prefixLen) : DataPage.getKey(cache, header[pos].dataAddr);
	}

	static byte[] getKey(PageCache cache, long chainIndex) throws IOException {
		ChainPage cp = RC_Page(cache, chainIndex);
		long dataAddr;
		cp.headerLock.lock();
		try {
			dataAddr = cp.header[cp.l_pos].dataAddr;
		} finally {
			cp.headerLock.unlock();
		}
		return DataPage.getKey(cache, dataAddr);
	}

	void insert(int i_pos, byte key[], long dataAddr) throws IOException {
		page.setDirty();
		ChainHDR hdr;
		if (chainSize() == CHAINCOUNT) {
			ChainPage nextPage;
			if (nextIndex == 0) {
				nextPage = cache.allocChainPage();
				nextPage.prevIndex = page.index;
				nextIndex = nextPage.page.index;
				if (i_pos == r_pos) {
					nextPage.insert(nextPage.l_pos, key, dataAddr);
					return;
				}
				r_pos--;
				nextPage.insert(nextPage.l_pos, getKey(r_pos),
						header[r_pos].dataAddr);
			} else {
				long lNextIndex = nextIndex;
				cache.lockChain(lNextIndex);
				nextPage = WC_Page(nextIndex);
				try {
					if (nextPage.chainSize() == CHAINCOUNT) {
						// avoid deeper recursion under control
						ChainPage splitPage = cache.allocChainPage();
						splitPage.nextIndex = nextPage.page.index;
						nextPage.prevIndex = splitPage.page.index;
						splitPage.prevIndex = page.index;
						nextIndex = splitPage.page.index;
						nextPage = splitPage;
					}
					if (i_pos == r_pos) {
						nextPage.insert(nextPage.l_pos, key, dataAddr);
						return;
					}
					r_pos--;
					nextPage.insert(nextPage.l_pos, getKey(r_pos),
							header[r_pos].dataAddr);
				} finally {
					cache.unlockChain(lNextIndex);
				}
			}
		}
		boolean writeLockIndex = false;
		try {
			if (i_pos == l_pos) {
				cache.wLockIndex();
				writeLockIndex = true;
				if (chainSize() > 0) {
					cache.updateChainIndex(getKey(l_pos), key, page.index);
				} else {
					cache.updateChainIndex(key, page.index);
				}
			}
			if (l_pos == 0 || ((l_pos + r_pos) >> 1) < i_pos
					&& r_pos < CHAINCOUNT) {
				hdr = header[r_pos];
				System.arraycopy(header, i_pos, header, i_pos + 1, r_pos
						- i_pos);
				r_pos++;
			} else {
				i_pos--;
				headerLock.lock();
				try {
					hdr = header[--l_pos];
					System.arraycopy(header, l_pos + 1, header, l_pos, i_pos
							- l_pos);
				} finally {
					headerLock.unlock();
				}
			}
			header[i_pos] = hdr;
			hdr.prefixLen = (byte) Math.min(key.length, hdr.prefixKey.length);
			for (short i = 0; i < hdr.prefixLen; i++)
				hdr.prefixKey[i] = key[i];
			hdr.dataAddr = dataAddr;
		} finally {
			if (writeLockIndex)
				cache.wUnlockIndex();
		}
	}

	private void merge(ChainPage nextPage) throws IOException {
		if (nextPage.chainSize() > 0) {
			boolean writeLockIndex = false;
			try {
				cache.wLockIndex();
				writeLockIndex = true;
				cache.removeChainIndex(nextPage.getKey(nextPage.l_pos));
				ChainPage borrowPage = ChainPage.alloc(cache, page);
				if (r_pos + nextPage.chainSize() > CHAINCOUNT) {
					ChainHDR[] borrowHeader = borrowPage.header;
					for (int src = l_pos, dst = 0; src < r_pos; src++, dst++) {
						ChainHDR hdr = header[src];
						header[src] = borrowHeader[dst];
						borrowHeader[dst] = hdr;
					}
					for (int src = l_pos, dst = 0; src < r_pos; src++, dst++) {
						ChainHDR hdr = header[dst];
						header[dst] = borrowHeader[dst];
						borrowHeader[dst] = hdr;
					}
					r_pos -= l_pos;
					l_pos = 0;
				}
				ChainHDR[] borrowHeader = borrowPage.header;
				ChainHDR[] nextHeader = nextPage.header;
				for (int src = nextPage.l_pos, dst = 0; src < nextPage.r_pos; src++, dst++) {
					ChainHDR hdr = nextHeader[src];
					nextHeader[src] = borrowHeader[dst];
					borrowHeader[dst] = hdr;
				}
				int size = nextPage.chainSize();
				for (int src = 0; src < size; src++, r_pos++) {
					ChainHDR hdr = borrowHeader[src];
					borrowHeader[src] = header[r_pos];
					header[r_pos] = hdr;
				}
				borrowPage.free();
			} finally {
				if (writeLockIndex)
					cache.wUnlockIndex();
			}
		}
		nextIndex = nextPage.nextIndex;
		if (nextIndex != 0) {
			cache.lockChain(nextIndex);
			try {
				WC_Page(nextIndex).prevIndex = page.index;
			} finally {
				cache.unlockChain(nextIndex);
			}
		}
		nextPage.page.setDirty();
		cache.freeChainPage(nextPage);
	}

	private void removeMaintain() throws IOException {
		long lNextIndex = nextIndex;
		cache.lockChain(lNextIndex);
		try {
			ChainPage nextPage = RC_Page(nextIndex);
			if (chainSize() + nextPage.chainSize() > CHAINCOUNT)
				return;
			page.setDirty();
			merge(nextPage);
		} finally {
			cache.unlockChain(lNextIndex);
		}
	}

	static void removeMaintain(PageCache cache, long index) throws IOException {
		cache.lockChain(index);
		try {
			if (cache.removeMaintain(index))
				RC_Page(cache, index).removeMaintain();
		} finally {
			cache.unlockChain(index);
		}
	}

	long remove(int pos) throws IOException {
		page.setDirty();
		boolean writeLockIndex = false;
		try {
			if (pos == l_pos) {
				cache.wLockIndex();
				writeLockIndex = true;
				if (l_pos + 1 < r_pos) {
					cache.updateChainIndex(getKey(l_pos), getKey(l_pos + 1),
							page.index);
				} else {
					cache.removeChainIndex(getKey(l_pos));
				}
			}
			ChainHDR hdr = header[pos];
			if (((l_pos + r_pos) >> 1) < pos) {
				--r_pos;
				System.arraycopy(header, pos + 1, header, pos, r_pos - pos);
				pos = r_pos;
			} else {
				System.arraycopy(header, l_pos, header, l_pos + 1, pos - l_pos);
				pos = l_pos++;
				headerLock.lock();
				try {
					// act as barrier with getKey
					// header[l_pos] == header[l_pos + 1]
				} finally {
					headerLock.unlock();
				}
			}
			header[pos] = hdr;
		} finally {
			if (writeLockIndex)
				cache.wUnlockIndex();
		}
		long lPrevIndex = 0;
		long lNextIndex = 0;
		try {
			if (l_pos == r_pos) {
				if (prevIndex == 0 && nextIndex != 0) {
					cache.lockChain(lNextIndex = nextIndex);
					WC_Page(nextIndex).prevIndex = 0;
					cache.freeChainPage(this);
					return 0;
				}
			}
			if (prevIndex != 0) {
				if (!cache.tryLockChain(prevIndex))
					return prevIndex;
				lPrevIndex = prevIndex;
				ChainPage prevPage = RC_Page(prevIndex);
				if (chainSize() + prevPage.chainSize() <= CHAINCOUNT) {
					prevPage.page.setDirty();
					prevPage.merge(this);
					return 0;
				}
			}
			if (nextIndex != 0) {
				cache.lockChain(lNextIndex = nextIndex);
				ChainPage nextPage = RC_Page(nextIndex);
				if (chainSize() + nextPage.chainSize() <= CHAINCOUNT) {
					nextPage.page.setDirty();
					merge(nextPage);
				}
			}
			return 0;
		} finally {
			cache.unlockChain(lPrevIndex);
			cache.unlockChain(lNextIndex);
		}
	}

	Cursor next(Cursor r) throws IOException {
		int pos = r.getChainPos();
		if (++pos < r_pos)
			return r.set(pos, header[pos].dataAddr);
		if (nextIndex == 0)
			return r.reset();
		if (r.set(nextIndex) != null)
			return r.leftMost();
		r.reset();
		return null;
	}

	Cursor prev(Cursor r) throws IOException {
		int pos = r.getChainPos();
		if (--pos >= l_pos)
			return r.set(pos, header[pos].dataAddr);
		if (prevIndex == 0)
			return r.reset();
		if (r.set(prevIndex) != null)
			return r.rightMost();
		r.reset();
		return null;
	}

	Cursor findNext(final byte[] key, Cursor r) throws IOException {
		int pos = Arrays.binarySearch(header, l_pos, r_pos, key);
		if (pos >= 0)
			return next(r.set(pos, 0));
		pos = -pos - 1;
		if (pos == r_pos)
			return next(r.set(pos, 0));
		return r.set(pos, header[pos].dataAddr);
	}

	Cursor findPrev(final byte[] key, Cursor r) throws IOException {
		int pos = Arrays.binarySearch(header, l_pos, r_pos, key);
		if (pos >= 0) {
			if (pos == l_pos)
				return prev(r.set(pos, 0));
			return r.set(pos - 1, header[pos - 1].dataAddr);
		}
		pos = -pos - 1;
		if (pos == l_pos)
			return prev(r.set(pos, 0));
		return r.set(pos - 1, header[pos - 1].dataAddr);
	}

	Cursor leftMost(Cursor r) {
		return r.set(l_pos, header[l_pos].dataAddr);
	}

	Cursor rightMost(Cursor r) {
		return r.set(r_pos - 1, header[r_pos - 1].dataAddr);
	}

	void saveDataAddr(int pos, long dataAddr) {
		header[pos].dataAddr = dataAddr;
		page.setDirty();
	}

	static int lockEstimate(long recordCount) {
		int pos = Arrays.binarySearch(primes, recordCount / CHAINCOUNT);
		if (pos < 0) {
			pos = -pos - 1;
			if (pos == primes.length)
				pos--;
		}
		return (int) primes[pos];
	}
}
