package limax.edb;

import java.io.IOException;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.nio.file.StandardOpenOption;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;

import limax.util.ConcurrentEnvironment;
import limax.util.MBeans;
import limax.util.Resource;
import limax.util.Trace;

public final class DataBase implements AutoCloseable {
	private final LabeledHash labeledHash;
	private final Map<String, PageCache> tables = new ConcurrentHashMap<>();
	private final Map<String, PageFile> files = new ConcurrentHashMap<>();
	private final Map<String, AtomicInteger> refCount = new ConcurrentHashMap<>();
	private final ReentrantLock backupLock = new ReentrantLock();
	private final Environment env;
	private final FileLock openLock;
	private final GlobalLogger logger;
	private final String schedulerName;
	private final ScheduledThreadPoolExecutor scheduler;
	private final Runnable checkPointTask = new Runnable() {
		@Override
		public void run() {
			checkpoint();
			if (env.getCheckpointPeriod() > 0 && scheduler.getQueue().isEmpty())
				scheduler.schedule(this, env.getCheckpointPeriod(),
						TimeUnit.MILLISECONDS);
		}
	};
	private final Resource mbean;

	public DataBase(final Environment env, final Path path) throws Exception {
		openLock = FileChannel.open(path.resolve("lock"),
				StandardOpenOption.CREATE, StandardOpenOption.WRITE).tryLock();
		if (openLock == null)
			throw new IllegalStateException("EDB <" + path.toAbsolutePath()
					+ "> is in use.");
		schedulerName = "limax.edb.checkpoint." + path.toString();
		scheduler = ConcurrentEnvironment.getInstance().newScheduledThreadPool(
				schedulerName, 1);
		this.env = env;
		this.labeledHash = new LabeledHash(env, this);
		ThreadContext ctx = ThreadContext.get();
		ctx.enter(labeledHash);
		env.setDataBase(this);
		Path dataDir = path.resolve("data");
		Path logDir = path.resolve("log");
		Files.createDirectories(dataDir);
		Files.createDirectories(logDir);
		logger = new GlobalLogger(env, logDir);
		try (DirectoryStream<Path> dStream = Files.newDirectoryStream(dataDir)) {
			for (Path file : dStream) {
				String tableName = file.getFileName().toString();
				PageFile pageFile = new PageFile(file);
				PageCache cache = new PageCache(pageFile);
				files.put(tableName, pageFile);
				logger.add(pageFile, cache);
				refCount.put(tableName, new AtomicInteger());
				tables.put(tableName, cache);
			}
		}
		logger.recover(Long.MAX_VALUE);
		env.setCacheLowest(tables.size() * Configure.MIN_CACHELOW);
		if (env.getCheckpointPeriod() > 0)
			scheduler.schedule(checkPointTask, env.getCheckpointPeriod(),
					TimeUnit.MILLISECONDS);
		ctx.leave();
		mbean = MBeans.register(MBeans.root(), env,
				"limax.edb:type=Environment." + path.toString());
	}

	void scheduleCheckpoint() {
		if (env.getCheckpointOnCacheFull() && scheduler.getQueue().isEmpty())
			scheduler.execute(checkPointTask);
	}

	public synchronized void addTable(String[] tableAdd) throws Exception {
		backupLock.lock();
		try {
			boolean needRotate = false;
			Path dataDir = logger.getLoggerDirectory().resolveSibling("data");
			ThreadContext ctx = ThreadContext.get();
			ctx.enter(labeledHash);
			for (String name : tableAdd) {
				String tableName = name.toLowerCase();
				if (!tables.containsKey(tableName)) {
					PageFile pageFile = new PageFile(dataDir.resolve(tableName));
					PageCache cache = new PageCache(pageFile);
					files.put(tableName, pageFile);
					logger.add(pageFile, cache);
					refCount.put(tableName, new AtomicInteger());
					tables.put(tableName, cache);
					needRotate = true;
				}
			}
			ctx.leave();
			if (needRotate) {
				env.setCacheLowest(tables.size() * Configure.MIN_CACHELOW);
				logger.forceRotate();
				checkpoint();
			}
		} finally {
			backupLock.unlock();
		}
	}

	public synchronized void removeTable(String[] tableRemove) throws Exception {
		backupLock.lock();
		try {
			boolean needRotate = false;
			Path dataDir = logger.getLoggerDirectory().resolveSibling("data");
			for (String name : tableRemove) {
				String tableName = name.toLowerCase();
				PageCache cache = tables.remove(tableName);
				if (cache != null) {
					AtomicInteger ref = refCount.remove(tableName);
					synchronized (ref) {
						while (ref.get() > 0)
							ref.wait();
					}
					PageFile pageFile = files.remove(tableName);
					logger.remove(pageFile);
					pageFile.close();
					Files.delete(dataDir.resolve(tableName));
					needRotate = true;
				}
			}
			if (needRotate) {
				env.setCacheLowest(tables.size() * Configure.MIN_CACHELOW);
				logger.forceRotate();
				checkpoint();
			}
		} finally {
			backupLock.unlock();
		}
	}

	@Override
	public synchronized void close() throws Exception {
		ConcurrentEnvironment.getInstance().shutdown(schedulerName);
		for (Iterator<Map.Entry<String, AtomicInteger>> it = refCount
				.entrySet().iterator(); it.hasNext();) {
			AtomicInteger ref = it.next().getValue();
			it.remove();
			synchronized (ref) {
				while (ref.get() > 0)
					try {
						ref.wait();
					} catch (InterruptedException e) {
					}
			}
		}
		checkpoint();
		for (PageFile pageFile : files.values()) {
			try {
				pageFile.close();
			} catch (IOException e) {
			}
		}
		logger.close();
		mbean.close();
		openLock.release();
	}

	LabeledHash getLabeledCache() {
		return labeledHash;
	}

	public synchronized void checkpoint() {
		try {
			_checkpoint();
		} catch (Exception e) {
			Trace.fatal("limax.edb.checkpoint", e);
			System.exit(-1);
		}
	}

	private void _checkpoint() throws Exception {
		Collection<PageCache> tableCollection = tables.values();
		tableCollection.forEach(cache -> cache.writeFreeze());
		for (PageCache cache : tableCollection)
			cache.waitWriteFrozen();
		Map<PageCache, List<PageLayout>> snapshots = labeledHash
				.createSnapshots();
		tableCollection.forEach(cache -> {
			cache.setSnapshots(snapshots.get(cache));
			cache.writeUnfreeze();
		});
		if (snapshots.isEmpty() && !logger.isForceRotate())
			return;
		logger.prepare();
		LabeledHash.releaseSnapshots(snapshots);
		logger.commit();
	}

	private PageCache requestTable(String tableName) {
		refCount.get(tableName).incrementAndGet();
		return tables.get(tableName);
	}

	private void returnTable(String tableName) {
		AtomicInteger ref = refCount.get(tableName);
		synchronized (ref) {
			if (ref.decrementAndGet() == 0)
				ref.notify();
		}
	}

	public boolean insert(String tableName, final byte[] key, final byte[] value)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.put(key, value, false);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void replace(String tableName, final byte[] key, final byte[] value)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				cache.put(key, value, true);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void remove(String tableName, final byte[] key) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				cache.remove(key, false);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public byte[] find(String tableName, final byte[] key) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.find(key);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public boolean exist(String tableName, final byte[] key) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.exist(key);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public long recordCount(String tableName) {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.recordCount();
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public byte[] firstKey(String tableName) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.firstKey();
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public byte[] lastKey(String tableName) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.lastKey();
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public byte[] nextKey(String tableName, final byte[] key)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.nextKey(key);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public byte[] prevKey(String tableName, final byte[] key)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enter(labeledHash);
			try {
				return cache.prevKey(key);
			} finally {
				ThreadContext.get().leave();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void walk(String tableName, Query query) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enterWalk(labeledHash);
			try {
				cache.walk(query);
			} finally {
				ThreadContext.get().leaveWalk();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void walk(String tableName, final byte[] key, Query query)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enterWalk(labeledHash);
			try {
				cache.walk(key, query);
			} finally {
				ThreadContext.get().leaveWalk();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void rwalk(String tableName, Query query) throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enterWalk(labeledHash);
			try {
				cache.rwalk(query);
			} finally {
				ThreadContext.get().leaveWalk();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void rwalk(String tableName, final byte[] key, Query query)
			throws IOException {
		PageCache cache = requestTable(tableName);
		try {
			ThreadContext.get().enterWalk(labeledHash);
			try {
				cache.rwalk(key, query);
			} finally {
				ThreadContext.get().leaveWalk();
			}
		} finally {
			returnTable(tableName);
		}
	}

	public void backup(Path path, boolean increment) throws IOException {
		if (!backupLock.tryLock())
			return;
		try {
			Path logDir = path.resolve("log");
			if (increment && logger.isSameBackupDirectory(logDir))
				return;
			Path dataDir = path.resolve("data");
			Files.createDirectories(dataDir);
			Files.createDirectories(logDir);
			logger.suspendBackupDirectory();
			try (DirectoryStream<Path> dStream = Files
					.newDirectoryStream(logger.getLoggerDirectory()
							.resolveSibling("data"))) {
				for (Path file : dStream)
					Files.copy(file, dataDir.resolve(file.getFileName()),
							StandardCopyOption.REPLACE_EXISTING,
							StandardCopyOption.COPY_ATTRIBUTES);
			}
			Set<Path> logSet = new HashSet<Path>();
			Set<Path> newSet = new HashSet<Path>();
			while (true) {
				try (DirectoryStream<Path> dStream = Files
						.newDirectoryStream(logger.getLoggerDirectory())) {
					for (Path file : dStream)
						newSet.add(file);
				}
				newSet.removeAll(logSet);
				if (newSet.size() == 0)
					break;
				for (Path file : newSet) {
					Files.copy(file, logDir.resolve(file.getFileName()),
							StandardCopyOption.REPLACE_EXISTING,
							StandardCopyOption.COPY_ATTRIBUTES);
					logSet.add(file);
				}
			}
			if (increment)
				logger.setBackupDirectory(logDir);
			else
				logger.resetBackupDirectory();
		} finally {
			backupLock.unlock();
		}
	}
}
