package limax.edb;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import limax.util.ConcurrentEnvironment;

final class GlobalLogger {
	private final static int posRecordStart = 0;
	private final static int posPageHeader = 4;
	private final static int posRecordNumber = 8;
	private final static int posChain = 12;

	private final Map<UUID, PageLogger> loggers = new HashMap<>();
	private final PageLayout keyPageLayout = new PageLayout();
	private final Environment env;
	private final Path loggerDirectory;
	private final String poolName;
	private final ExecutorService pool;
	private PageFile loggerFile;
	private PageFile prevLoggerFile;
	private Path backupDirectory = null;
	private boolean forceRotate = false;

	private long loggerId;

	GlobalLogger(Environment env, Path loggerDirectory) {
		this.env = env;
		this.loggerDirectory = loggerDirectory;
		poolName = "limax.edb.GlobalLogger."
				+ loggerDirectory.toString().replace(':', '_');
		pool = ConcurrentEnvironment.getInstance().newThreadPool(poolName, 0);
	}

	void close() throws IOException {
		ConcurrentEnvironment.getInstance().shutdown(poolName);
		for (PageLogger logger : loggers.values())
			logger.close();
		keyPageLayout.free();
		loggerFile.close();
	}

	void add(PageFile pageFile, PageCache cache) throws IOException {
		loggers.put(pageFile.getUUID(), new PageLogger(pageFile, cache));
	}

	void remove(PageFile pageFile) {
		loggers.remove(pageFile.getUUID()).close();
	}

	private static int recordPosition(int recordNumber) {
		return Configure.LOGGER_PREFIX_LENGTH + Configure.LOGGER_HEADER_LENGTH
				* recordNumber;
	}

	private int getRecordStart() {
		return keyPageLayout.data().getInt(posRecordStart);
	}

	private int getPageHeader() {
		return keyPageLayout.data().getInt(posPageHeader);
	}

	private int getRecordNumber() {
		return keyPageLayout.data().getInt(posRecordNumber);
	}

	private long getChain() {
		return keyPageLayout.data().getLong(posChain);
	}

	private long getTimeStamp(int recordNumber) {
		return keyPageLayout.data().getLong(recordPosition(recordNumber));
	}

	private long getTimeStamp() {
		return getTimeStamp(getRecordNumber());
	}

	private long nearestLoggerId(long loggerLastCheck) throws IOException {
		long nearestLoggerId = Long.MIN_VALUE;
		try (DirectoryStream<Path> dStream = Files
				.newDirectoryStream(loggerDirectory)) {
			for (Path file : dStream) {
				long id = Long.valueOf(file.getFileName().toString(), 16);
				if (id > nearestLoggerId && id <= loggerLastCheck)
					nearestLoggerId = id;
			}
		}
		return nearestLoggerId;
	}

	private void load(long newLoggerId) throws IOException {
		if (loggerFile != null)
			loggerFile.close();
		Path loggerFileName = loggerDirectory.resolve(String.format("%016x",
				loggerId = newLoggerId));
		loggerFile = new PageFile(loggerFileName);
		loggerFile.read(0, keyPageLayout);
		int recordStart = getRecordStart();
		byte uuid[] = new byte[36];
		PageLayout layout = new PageLayout();
		for (int i = 1; i < recordStart; i++) {
			loggerFile.read(i, layout);
			layout.data().get(uuid).rewind();
			PageLogger pageLogger = loggers.get(UUID
					.fromString(new String(uuid)));
			pageLogger.getKeyPageLayout().data().clear();
			pageLogger.getKeyPageLayout().data().put(layout.data());
			pageLogger.load(loggerFile, i);
			layout.data().clear();
		}
		layout.free();
	}

	private int checkRotate() throws IOException {
		int pageHeader = getPageHeader();
		if (pageHeader > env.getLoggerPages()
				|| getRecordNumber() == Configure.LOGGER_HEADER_MAX
				|| forceRotate)
			pageHeader = rotate();
		return pageHeader;
	}

	private synchronized int rotate() throws IOException {
		long newLoggerId = System.currentTimeMillis();
		loggerId = loggerId < newLoggerId ? newLoggerId : loggerId + 1;
		prevLoggerFile = loggerFile;
		loggerFile = new PageFile(loggerDirectory.resolve(String.format(
				"%016x", loggerId)));
		PageLayout prevPageLayout = keyPageLayout.createSnapshot();
		int h = loggers.size() + 1;
		keyPageLayout.zero().putInt(h).putInt(h).putInt(0).putLong(0);
		int pageIndex = 1;
		for (PageLogger logger : loggers.values())
			logger.rotate(loggerFile, pageIndex++);
		loggerFile.write(0, keyPageLayout);
		loggerFile.sync();
		if (prevLoggerFile != null) {
			prevPageLayout.data().putLong(posChain, loggerId);
			prevLoggerFile.write(0, prevPageLayout);
			prevLoggerFile.close();
		}
		prevPageLayout.free();
		return h;
	}

	void prepare() throws IOException {
		int pageIndex = checkRotate();
		int recordNumber = getRecordNumber();
		for (PageLogger logger : loggers.values())
			pageIndex = logger.prepare(pageIndex, recordNumber);
		loggerFile.sync();
		keyPageLayout.data().putLong(recordPosition(recordNumber),
				System.currentTimeMillis());
		loggerFile.write(0, keyPageLayout);
		loggerFile.sync();
	}

	private void wait(List<Future<Void>> results) throws IOException {
		for (Future<Void> r : results) {
			try {
				r.get();
			} catch (InterruptedException | ExecutionException e) {
				throw new IOException(e.getCause());
			}
		}
	}

	void commit() throws IOException {
		final int recordNumber = getRecordNumber();
		final long timeStamp = getTimeStamp(recordNumber);
		List<Future<Void>> results = new ArrayList<>();
		for (final PageLogger logger : loggers.values())
			results.add(pool.submit(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					logger.commit(loggerId, timeStamp, recordNumber);
					return null;
				}
			}));
		int pageHeader = getPageHeader();
		for (PageLogger logger : loggers.values())
			pageHeader = Math.max(pageHeader,
					logger.lastPageIndex(recordNumber));
		keyPageLayout.data().putInt(posPageHeader, pageHeader);
		keyPageLayout.data().putInt(posRecordNumber, recordNumber + 1);
		wait(results);
		loggerFile.write(0, keyPageLayout);
		loggerFile.sync();
		rotateCleanup();
	}

	private void truncateLogger() throws IOException {
		rotate();
		commit(); // create new null checkpoint
		suspendBackupDirectory();
		resetBackupDirectory();
	}

	List<Long> getValidLoggerId() throws IOException {
		List<Long> timeStamps = new ArrayList<>();
		long maxLoggerLastCheck = Long.MIN_VALUE;
		for (PageLogger logger : loggers.values()) {
			long loggerLastCheck = logger.getLoggerLastCheck();
			if (loggerLastCheck == 0)
				throw new RuntimeException("Not Valid Backup with new table <"
						+ logger.getTableName() + ">");
			maxLoggerLastCheck = Math.max(maxLoggerLastCheck, loggerLastCheck);
		}
		load(nearestLoggerId(maxLoggerLastCheck));
		int totalRecordNumber = getRecordNumber();
		int recordNumber = 0;
		for (; recordNumber < totalRecordNumber
				&& getTimeStamp(recordNumber) < maxLoggerLastCheck; recordNumber++)
			;
		while (true) {
			for (; recordNumber < totalRecordNumber; recordNumber++)
				timeStamps.add(getTimeStamp(recordNumber));
			long chain = getChain();
			if (chain == 0
					|| !Files.exists(loggerDirectory.resolve(String.format(
							"%016x", chain))))
				break;
			load(chain);
			recordNumber = 0;
			totalRecordNumber = getRecordNumber();
		}
		return timeStamps;
	}

	private class MagicLogger {
		private Map<PageLogger, Entry> map = new IdentityHashMap<>();

		private class Entry {
			private long loggerId;
			private long loggerLastCheck;

			Entry(PageLogger logger) {
				this.loggerId = logger.getLoggerId();
				this.loggerLastCheck = logger.getLoggerLastCheck();
			}

			void commit(PageLogger logger) throws IOException {
				logger.commit(loggerId, loggerLastCheck);
			}
		}

		void commit() throws IOException {
			List<Future<Void>> results = new ArrayList<>();
			for (final Map.Entry<PageLogger, Entry> e : map.entrySet()) {
				results.add(pool.submit(new Callable<Void>() {
					@Override
					public Void call() throws Exception {
						e.getValue().commit(e.getKey());
						return null;
					}
				}));
			}
			GlobalLogger.this.wait(results);
		}

		void commit(PageLogger logger, int recordNumber) throws IOException {
			if (!map.containsKey(logger))
				map.put(logger, new Entry(logger));
			logger.commit(recordNumber);
		}
	}

	void recover(long until) throws IOException {
		if (loggers.size() == 0)
			return;
		long minLoggerLastCheck = Long.MAX_VALUE;
		for (PageLogger logger : loggers.values())
			minLoggerLastCheck = Math.min(minLoggerLastCheck,
					logger.getLoggerLastCheck());
		load(nearestLoggerId(minLoggerLastCheck));
		int totalRecordNumber = getRecordNumber();
		int recordNumber = 0;
		MagicLogger magicLogger = new MagicLogger();
		for (; recordNumber < totalRecordNumber
				&& getTimeStamp(recordNumber) <= minLoggerLastCheck; recordNumber++)
			;
		long timeStamp = 0;
		while (true) {
			for (; recordNumber < totalRecordNumber
					&& (timeStamp = getTimeStamp(recordNumber)) <= until; recordNumber++) {
				for (PageLogger logger : loggers.values())
					if (logger.getLoggerLastCheck() < timeStamp)
						magicLogger.commit(logger, recordNumber);
			}
			if (timeStamp >= until) {
				magicLogger.commit();
				truncateLogger(); // for restore need not reloadPageCache
				return;
			}
			long chain = getChain();
			if (chain == 0)
				break;
			load(chain);
			recordNumber = 0;
			totalRecordNumber = getRecordNumber();
		}
		if (getTimeStamp() != 0) {
			commit();
			for (PageLogger logger : loggers.values())
				logger.reloadPageCache();
		} else {
			magicLogger.commit();
		}
	}

	Path getLoggerDirectory() {
		return loggerDirectory;
	}

	void forceRotate() {
		forceRotate = true;
	}

	boolean isForceRotate() {
		return forceRotate;
	}

	synchronized boolean isSameBackupDirectory(Path path) {
		try {
			return Files.isSameFile(path, backupDirectory);
		} catch (IOException e) {
			return false;
		}
	}

	synchronized void suspendBackupDirectory() {
		backupDirectory = null;
	}

	synchronized void resetBackupDirectory() throws IOException {
		try (DirectoryStream<Path> dStream = Files
				.newDirectoryStream(loggerDirectory)) {
			for (Path file : dStream) {
				if (file.equals(loggerFile.getPath()) || prevLoggerFile != null
						&& file.equals(prevLoggerFile.getPath()))
					continue;
				Files.delete(file);
			}
		}
	}

	synchronized void setBackupDirectory(Path path) throws IOException {
		try (DirectoryStream<Path> dStream = Files
				.newDirectoryStream(loggerDirectory)) {
			for (Path file : dStream) {
				if (file.equals(loggerFile.getPath()) || prevLoggerFile != null
						&& file.equals(prevLoggerFile.getPath()))
					continue;
				Files.move(file, backupDirectory.resolve(file.getFileName()),
						StandardCopyOption.REPLACE_EXISTING);
			}
		}
		backupDirectory = path;
	}

	private synchronized void rotateCleanup() throws IOException {
		if (prevLoggerFile == null)
			return;
		if (backupDirectory == null) {
			prevLoggerFile.delete();
		} else {
			Files.move(prevLoggerFile.getPath(), backupDirectory
					.resolve(prevLoggerFile.getPath().getFileName()),
					StandardCopyOption.REPLACE_EXISTING);
		}
		prevLoggerFile = null;
		if (forceRotate) {
			forceRotate = false;
			suspendBackupDirectory();
			resetBackupDirectory();
		}
	}
}
