package limax.edb;

import java.io.IOException;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

final class PageCache extends PageLoader {
	final private ConcurrentHashMap<Long, Long> removeMaintain = new ConcurrentHashMap<>();
	final private static HashLock loadLock = new HashLock(53);
	final private ReentrantLock allocLock = new ReentrantLock();
	final private ReentrantLock fragmentLock = new ReentrantLock();
	final private ReentrantReadWriteLock indexLock = new ReentrantReadWriteLock(
			true);
	final private ReentrantLock barrierLock = new ReentrantLock();
	final private Condition snapshotCondition = barrierLock.newCondition();
	final private Condition writeOpCondition = barrierLock.newCondition();
	private boolean snapshotBarrier = false;
	private int writeOpCount = 0;
	private List<PageLayout> snapshots;
	private HashLock chainLock;
	private MagicPage magic;

	PageCache(PageFile pageFile) throws IOException {
		super(pageFile);
		if (pageFile.size() == 0) {
			magic = MagicPage.alloc();
			allocRootIndexPage();
		} else {
			magic = MagicPage.alloc(pageFile);
		}
		this.chainLock = new HashLock(ChainPage.lockEstimate(recordCount()));
	}

	private void waitWritePermit() throws IOException {
		barrierLock.lock();
		try {
			while (snapshotBarrier)
				snapshotCondition.await();
			writeOpCount++;
		} catch (InterruptedException e) {
			throw new IOException(e.getCause());
		} finally {
			barrierLock.unlock();
		}
	}

	private void releaseWritePermit() {
		barrierLock.lock();
		try {
			writeOpCount--;
			writeOpCondition.signal();
		} finally {
			barrierLock.unlock();
		}
	}

	void writeFreeze() {
		barrierLock.lock();
		try {
			snapshotBarrier = true;
		} finally {
			barrierLock.unlock();
		}
	}

	void waitWriteFrozen() throws InterruptedException {
		barrierLock.lock();
		try {
			while (writeOpCount > 0)
				writeOpCondition.await();
		} finally {
			barrierLock.unlock();
		}
	}

	void writeUnfreeze() {
		barrierLock.lock();
		try {
			snapshotBarrier = false;
			snapshotCondition.signalAll();
		} finally {
			barrierLock.unlock();
		}
	}

	void setSnapshots(List<PageLayout> snapshots) {
		if (snapshots != null)
			snapshots.add(magic.createSnapshot());
		this.snapshots = snapshots;
	}

	List<PageLayout> getSnapshots() {
		return snapshots;
	}

	MagicPage getMagicPage() {
		return magic;
	}

	void lockAlloc() {
		allocLock.lock();
	}

	void unlockAlloc() {
		allocLock.unlock();
	}

	void lockFragment() {
		fragmentLock.lock();
	}

	void unlockFragment() {
		fragmentLock.unlock();
	}

	@Override
	protected Page loadPage(long index) throws IOException {
		ThreadContext ctx = ThreadContext.get();
		Page page = ctx.get(index, this);
		if (page == null) {
			loadLock.lock(index);
			try {
				page = ctx.get(index, this);
				if (page == null) {
					page = Page.alloc(index, this, pageFile);
					ctx.put(page);
				}
			} finally {
				loadLock.unlock(index);
			}
		}
		return page;
	}

	private Page allocPage() throws IOException {
		allocLock.lock();
		Page page;
		long index = magic.getFreePageList();
		if (index == 0) {
			index = magic.extendPage();
			allocLock.unlock();
			page = Page.alloc(index, this);
			ThreadContext.get().put(page);
		} else {
			try {
				page = loadPage(index);
				magic.setFreePageList(page.asFreePage().getFreePageList());
			} finally {
				allocLock.unlock();
			}
		}
		return page.setDirty();
	}

	DataPage allocDataPage() throws IOException {
		return allocPage().asDataPage();
	}

	IndexPage allocIndexPage() throws IOException {
		return allocPage().asIndexPage();
	}

	IndexPage allocRootIndexPage() throws IOException {
		Page page = allocPage();
		magic.setRootIndexIndex(page.index);
		return page.asIndexPage();
	}

	ChainPage allocChainPage() throws IOException {
		return allocPage().asChainPage();
	}

	void setRootIndex(Page rootPage) {
		magic.setRootIndexIndex(rootPage.index);
	}

	IndexPage getRootIndex() throws IOException {
		return loadPage(magic.getRootIndexIndex()).asIndexPage();
	}

	private Cursor allocCursor() {
		return new Cursor(this);
	}

	void lockChain(long index) {
		if (index != 0)
			chainLock.lock(index);
	}

	boolean tryLockChain(long index) {
		return index == 0 ? true : chainLock.tryLock(index);
	}

	void unlockChain(long index) {
		if (index != 0)
			chainLock.unlock(index);
	}

	void rLockIndex() {
		indexLock.readLock().lock();
	}

	void rUnlockIndex() {
		indexLock.readLock().unlock();
	}

	void wLockIndex() {
		indexLock.writeLock().lock();
	}

	void wUnlockIndex() {
		indexLock.writeLock().unlock();
	}

	void removeChainIndex(final byte[] oldKey) throws IOException {
		getRootIndex().removeChainIndex(oldKey);
	}

	void updateChainIndex(final byte[] newKey, long chainIndex)
			throws IOException {
		getRootIndex().insertChainIndex(newKey, chainIndex);
	}

	void updateChainIndex(final byte[] oldKey, final byte[] newKey,
			long chainIndex) throws IOException {
		IndexPage rootIndexPage = getRootIndex();
		rootIndexPage.removeChainIndex(oldKey);
		rootIndexPage.insertChainIndex(newKey, chainIndex);
	}

	private Cursor leftMost(Cursor r) throws IOException {
		while (true) {
			rLockIndex();
			try {
				if (getRootIndex().leftMost(r) != null)
					return r.hasChain() ? r.leftMost() : null;
			} finally {
				rUnlockIndex();
			}
		}
	}

	private Cursor rightMost(Cursor r) throws IOException {
		while (true) {
			rLockIndex();
			try {
				if (getRootIndex().rightMost(r) != null)
					return r.hasChain() ? r.rightMost() : null;
			} finally {
				rUnlockIndex();
			}
		}
	}

	private boolean findIndex(final byte[] key, Cursor r) throws IOException {
		while (true) {
			rLockIndex();
			try {
				if (getRootIndex().find(key, r) != null)
					return r.hasChain();
			} finally {
				rUnlockIndex();
			}
		}
	}

	private Cursor findKey(final byte[] key, Cursor r) throws IOException {
		if (!findIndex(key, r))
			return null;
		if (!r.find(key).isFound())
			return null;
		return r;
	}

	boolean put(final byte[] key, final byte[] value, boolean replace)
			throws IOException {
		waitWritePermit();
		try {
			Cursor r = allocCursor();
			if (!findIndex(key, r)) {
				wLockIndex();
				try {
					if (!findIndex(key, r)) {
						allocChainPage().insert(0, key,
								DataPage.createDataNode(this, key, value));
						magic.incrementRecordCount();
						return true;
					}
				} finally {
					wUnlockIndex();
				}
			}
			try {
				r.find(key);
				if (r.isFound()) {
					if (!replace)
						return false;
					long dataAddr = DataPage.replaceDataNode(this, key, value,
							r.getDataAddr());
					if (dataAddr != r.getDataAddr())
						r.saveDataAddr(dataAddr);
				} else {
					r.insert(key, DataPage.createDataNode(this, key, value));
					magic.incrementRecordCount();
				}
			} finally {
				r.unlockChain();
			}
			return true;
		} finally {
			releaseWritePermit();
		}
	}

	byte[] find(final byte[] key) throws IOException {
		Cursor r = allocCursor();
		try {
			return findKey(key, r) != null ? DataPage.getValue(this,
					r.getDataAddr()) : null;
		} finally {
			r.unlockChain();
		}
	}

	boolean exist(final byte[] key) throws IOException {
		Cursor r = allocCursor();
		try {
			return findKey(key, r) != null;
		} finally {
			r.unlockChain();
		}
	}

	long recordCount() {
		return magic.getRecordCount();
	}

	void freeChainPage(ChainPage page) {
		removeMaintain.remove(page.getPage().index);
		page.getPage().asFreePage();
	}

	boolean removeMaintain(long index) {
		return removeMaintain.remove(index, index);
	}

	byte[] remove(final byte[] key, boolean needOldValue) throws IOException {
		waitWritePermit();
		try {
			Cursor r = allocCursor();
			long index = 0;
			try {
				if (findKey(key, r) == null)
					return null;
				index = r.remove();
				if (index != 0)
					removeMaintain.put(index, index);
				byte[] value = null;
				if (needOldValue)
					value = DataPage.getValue(this, r.getDataAddr());
				DataPage.freeDataNode(this, r.getDataAddr());
				magic.decrementRecordCount();
				return value;
			} finally {
				r.unlockChain();
				if (index != 0)
					ChainPage.removeMaintain(this, index);
			}
		} finally {
			releaseWritePermit();
		}
	}

	byte[] firstKey() throws IOException {
		Cursor r = allocCursor();
		try {
			return leftMost(r) == null ? null : r.getKey();
		} finally {
			r.unlockChain();
		}
	}

	byte[] lastKey() throws IOException {
		Cursor r = allocCursor();
		try {
			return rightMost(r) == null ? null : r.getKey();
		} finally {
			r.unlockChain();
		}
	}

	byte[] nextKey(final byte[] key) throws IOException {
		Cursor r = allocCursor();
		try {
			while (true) {
				if (!findIndex(key, r))
					return null;
				if (r.findNext(key) == null)
					continue;
				return r.hasChain() ? r.getKey() : null;
			}
		} finally {
			r.unlockChain();
		}
	}

	byte[] prevKey(final byte[] key) throws IOException {
		Cursor r = allocCursor();
		try {
			while (true) {
				if (!findIndex(key, r))
					return null;
				if (r.findPrev(key) == null)
					continue;
				return r.hasChain() ? r.getKey() : null;
			}
		} finally {
			r.unlockChain();
		}
	}

	private byte[] queryUpdate(Cursor r, Query query) throws IOException {
		if (query instanceof QueryKey) {
			byte key[] = r.getKey();
			return ((QueryKey) query).update(key) ? key : null;
		}
		byte data[][] = DataPage.getData(this, r.getDataAddr());
		return ((QueryData) query).update(data[0], data[1]) ? data[0] : null;
	}

	private byte[] walk(Cursor r, Query query) throws IOException {
		ThreadContext ctx = ThreadContext.get();
		while (true) {
			byte[] key = queryUpdate(r, query);
			if (key == null)
				break;
			if (r.next() == null)
				return key;
			if (!r.isFound())
				break;
			ctx.cleanupWalk(r.getPage());
		}
		return null;
	}

	private void walk(byte[] key, Query query, Cursor r) throws IOException {
		while (true) {
			if (!findIndex(key, r))
				return;
			if (r.findNext(key) == null)
				continue;
			if ((key = walk(r, query)) == null)
				return;
		}
	}

	void walk(final byte[] key, Query query) throws IOException {
		Cursor r = allocCursor();
		try {
			walk(key, query, r);
		} finally {
			r.unlockChain();
		}
	}

	void walk(Query query) throws IOException {
		Cursor r = allocCursor();
		try {
			if (leftMost(r) == null)
				return;
			byte[] key = walk(r, query);
			if (key != null)
				walk(key, query, r);
		} finally {
			r.unlockChain();
		}
	}

	private byte[] rwalk(Cursor r, Query query) throws IOException {
		ThreadContext ctx = ThreadContext.get();
		while (true) {
			byte[] key = queryUpdate(r, query);
			if (key == null)
				break;
			if (r.prev() == null)
				return key;
			if (!r.isFound())
				break;
			ctx.cleanupWalk(r.getPage());
		}
		return null;
	}

	private void rwalk(byte[] key, Query query, Cursor r) throws IOException {
		while (true) {
			if (!findIndex(key, r))
				return;
			if (r.findPrev(key) == null)
				continue;
			if ((key = rwalk(r, query)) == null)
				return;
		}
	}

	void rwalk(final byte[] key, Query query) throws IOException {
		Cursor r = allocCursor();
		try {
			rwalk(key, query, r);
		} finally {
			r.unlockChain();
		}
	}

	void rwalk(Query query) throws IOException {
		Cursor r = allocCursor();
		try {
			if (rightMost(r) == null)
				return;
			byte[] key = rwalk(r, query);
			if (key != null)
				rwalk(key, query, r);
		} finally {
			r.unlockChain();
		}
	}
}
