package limax.edb;

import java.io.IOException;
import java.nio.channels.ClosedByInterruptException;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.EnumSet;
import java.util.Set;
import java.util.UUID;

final class PageFile {
	final static private Set<StandardOpenOption> options = EnumSet.of(
			StandardOpenOption.CREATE, StandardOpenOption.READ,
			StandardOpenOption.WRITE);
	private FileChannel channel;
	final private Path path;
	final private UUID uuid;

	PageFile(Path path) throws IOException {
		this.path = path;
		this.channel = FileChannel.open(path, options);
		this.uuid = UUID.nameUUIDFromBytes(path.getFileName().toString()
				.getBytes());
	}

	void read(long index, PageLayout layout) throws IOException {
		layout.data().clear();
		int n = 0;
		for (;;) {
			try {
				n = channel.read(layout.data(), index * Configure.PAGESIZE);
			} catch (ClosedByInterruptException e) {
				channel = FileChannel.open(path, options);
				Thread.interrupted();
				continue;
			}
			break;
		}
		if (n != Configure.PAGESIZE)
			throw new IOException(path + " read page " + index + " bad size = "
					+ n);
		layout.data().flip();
	}

	void write(long index, PageLayout layout) throws IOException {
		layout.data().rewind();
		int n = 0;
		for (;;) {
			try {
				n = channel.write(layout.data(), index * Configure.PAGESIZE);
			} catch (ClosedByInterruptException e) {
				channel = FileChannel.open(path, options);
				Thread.interrupted();
				continue;
			}
			break;
		}
		if (n != Configure.PAGESIZE)
			throw new IOException(path + " write page " + index
					+ " bad size = " + n);
	}

	void sync() throws IOException {
		channel.force(false);
	}

	long size() throws IOException {
		return channel.size();
	}

	UUID getUUID() {
		return uuid;
	}

	Path getPath() {
		return path;
	}

	void close() throws IOException {
		channel.close();
	}

	void delete() throws IOException {
		Files.delete(path);
	}
}
