package limax.edb;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.List;

final class PageLogger {
	final private PageFile pageFile;
	final private PageCache cache;
	final private PageLayout keyPageLayout;
	final private MagicPage magic;

	private PageFile loggerFile;
	private int keyPageIndex;

	PageLogger(PageFile pageFile, PageCache cache) throws IOException {
		this.cache = cache;
		this.pageFile = pageFile;
		this.keyPageLayout = new PageLayout();
		this.magic = pageFile.size() > 0 ? MagicPage.alloc(pageFile)
				: MagicPage.alloc();
	}

	void close() {
		keyPageLayout.free();
		magic.free();
	}

	private static int recordPosition(int recordNumber) {
		return Configure.LOGGER_PREFIX_LENGTH + Configure.LOGGER_HEADER_LENGTH
				* recordNumber;
	}

	PageLayout getKeyPageLayout() {
		return keyPageLayout;
	}

	void load(PageFile loggerFile, int keyPageIndex) {
		this.loggerFile = loggerFile;
		this.keyPageIndex = keyPageIndex;
	}

	void rotate(PageFile loggerFile, int keyPageIndex) throws IOException {
		keyPageLayout.zero().put(pageFile.getUUID().toString().getBytes());
		this.loggerFile = loggerFile;
		this.keyPageIndex = keyPageIndex;
		loggerFile.write(keyPageIndex, keyPageLayout);
	}

	int prepare(int pageIndex, int recordNumber) throws IOException {
		ByteBuffer bb = keyPageLayout.data();
		List<PageLayout> snapshots = cache.getSnapshots();
		bb.position(recordPosition(recordNumber));
		bb.putInt(pageIndex);
		if (snapshots != null)
			for (PageLayout snapshot : snapshots)
				loggerFile.write(pageIndex++, snapshot);
		bb.putInt(pageIndex);
		loggerFile.write(keyPageIndex, keyPageLayout);
		return pageIndex;
	}

	int lastPageIndex(int recordNumber) {
		return keyPageLayout.data().getInt(recordPosition(recordNumber) + 4);
	}

	void commit(int recordNumber) throws IOException {
		ByteBuffer bb = keyPageLayout.data();
		bb.position(recordPosition(recordNumber));
		int it = bb.getInt();
		int ie = bb.getInt() - 1;
		if (it > ie)
			return;
		PageLayout layout = new PageLayout();
		for (; it != ie; it++) {
			loggerFile.read(it, layout);
			pageFile.write(
					Configure.addr2index(layout.data().getLong(
							Configure.PAGESIZE - 8)), layout);
		}
		magic.load(loggerFile, it);
		layout.free();
	}

	void commit(long loggerId, long timeStamp, int recordNumber)
			throws IOException {
		commit(recordNumber);
		magic.setLoggerId(loggerId);
		magic.setLoggerLastCheck(timeStamp);
		magic.save(pageFile);
		pageFile.sync();
	}

	void commit(long oldLoggerId, long oldTimeStamp) throws IOException {
		long newLoggerId = magic.getLoggerId();
		long newTimeStamp = magic.getLoggerLastCheck();
		magic.setLoggerId(oldLoggerId);
		magic.setLoggerLastCheck(oldTimeStamp);
		magic.save(pageFile);
		pageFile.sync();
		magic.setLoggerId(newLoggerId);
		magic.setLoggerLastCheck(newTimeStamp);
		magic.save(pageFile);
		pageFile.sync();
		reloadPageCache();
	}

	long getLoggerId() {
		return cache.getMagicPage().getLoggerId();
	}

	long getLoggerLastCheck() {
		return cache.getMagicPage().getLoggerLastCheck();
	}

	String getTableName() {
		return pageFile.getPath().getFileName().toString();
	}

	void reloadPageCache() throws IOException {
		cache.getMagicPage().load(pageFile, 0);
	}
}
