package limax.endpoint;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

import limax.defines.ErrorSource;
import limax.defines.SessionType;
import limax.net.Engine;
import limax.util.ConcurrentEnvironment;
import limax.util.Dispatcher.Dispatchable;

public final class Endpoint {

	private Endpoint() {
	}

	public static void openEngine(int netProcessors, int protocolSchedulers,
			int applicationExecutors) throws Exception {
		Engine.open(netProcessors, protocolSchedulers, applicationExecutors);
	}

	public static void openEngine() throws Exception {
		openEngine(1, 1, 4);
	}

	public static void closeEngine(final Runnable done) {
		ExpansionImpl.cleanup();
		ConcurrentEnvironment.executeStandaloneTask(new Runnable() {
			@Override
			public void run() {
				try {
					Engine.close();
				} finally {
					if (done != null)
						done.run();
				}
			}
		});
	}

	public static EndpointConfigBuilder createLoginConfigBuilder(
			String serverIp, int serverPort, String username, String token,
			String platflag) {
		final EndpointConfigBuilderImpl config = new EndpointConfigBuilderImpl(
				serverIp, serverPort, platflag, username, token, false);
		return config.isConfigCanLogin() ? config : null;
	}

	public static EndpointConfigBuilder createPingOnlyConfigBuilder(
			String serverIp, int serverPort) {
		return new EndpointConfigBuilderImpl(serverIp, serverPort, "", "", "",
				true);
	}

	private static void mapPvidsAppendValue(Map<Short, Byte> pvids, Short s,
			int nv) {
		Byte v = pvids.get(s);
		if (null == v)
			v = Byte.valueOf((byte) nv);
		else
			v = Byte.valueOf((byte) (v | nv));
		pvids.put(s, v);
	}

	private static Collection<Short> makeProtocolProviderIds(
			EndpointConfig config) {
		final Set<Short> pvids = new HashSet<Short>();
		for (int type : config.getEndpointState().getSizePolicy().keySet()) {
			int pvid = type >> 16;
			if (pvid > 0)
				pvids.add((short) pvid);
		}
		return pvids;
	}

	private static Collection<Short> makeScriptProviderIds(EndpointConfig config) {
		if (config.getScriptEngineHandle() != null)
			return config.getScriptEngineHandle().getProviders();
		return Collections.emptySet();
	}

	private static Map<Short, Byte> makeProviderMap(EndpointConfig config) {
		final Map<Short, Byte> pvids = new HashMap<Short, Byte>();
		for (Short s : makeProtocolProviderIds(config))
			mapPvidsAppendValue(pvids, s, SessionType.ST_PROTOCOL);
		for (Short s : config.getStaticViewClasses().keySet())
			mapPvidsAppendValue(pvids, s, SessionType.ST_STATIC);
		for (Short s : config.getVariantProviderIds())
			mapPvidsAppendValue(pvids, s, SessionType.ST_VARIANT);
		for (Short s : makeScriptProviderIds(config))
			mapPvidsAppendValue(pvids, s, SessionType.ST_SCRIPT);
		mapPvidsAppendValue(pvids, limax.endpoint.auanyviews.ViewManager
				.getInstance().getProviderId(), SessionType.ST_STATIC);
		return pvids;
	}

	public static void start(final EndpointConfig config,
			final EndpointListener listener) {
		ConcurrentEnvironment.executeStandaloneTask(new Runnable() {
			@Override
			public void run() {
				try {
					if (null == Engine.getProtocolScheduler())
						throw new Exception("endpoint need call openEngine");
					Map<Short, Byte> pvids = makeProviderMap(config);
					if (!config.isPingServerOnly() && pvids.isEmpty())
						throw new Exception("endpoint no available provider");
					new EndpointManagerImpl(config, listener, pvids);
				} catch (final Exception e) {
					config.getClientManagerConfig().getDispatcher()
							.run(new Dispatchable() {
								@Override
								public void run() {
									listener.onErrorOccured(
											ErrorSource.ENDPOINT, 0, e);
								}
							});
				}
			}
		});
	}

	private static AtomicReference<EndpointManagerImpl> defaultEndpointManager = new AtomicReference<EndpointManagerImpl>();

	static void setDefaultEndpointManager(EndpointManagerImpl manager) {
		defaultEndpointManager.compareAndSet(null, manager);
	}

	static void clearDefaultEndpointManager(EndpointManagerImpl manager) {
		defaultEndpointManager.compareAndSet(manager, null);
	}

	public static EndpointManager getDefaultEndpointManager() {
		return defaultEndpointManager.get();
	}
}
