package limax.net;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import limax.net.io.FixedCpuPoll;
import limax.net.io.NetModel;
import limax.util.ConcurrentEnvironment;
import limax.util.HashExecutor;

public class Engine {
	private final static Object dummy = new Object();
	private final static int limitProtocolSize = Integer.getInteger(
			"limax.net.Engine.limitProtocolSize", 1048576).intValue();
	private volatile static ScheduledExecutorService protocolScheduler;
	private volatile static HashExecutor protocolExecutor;
	private volatile static HashExecutor applicationExecutor;
	private volatile static ThreadPoolExecutor engineExecutor;
	private final static Map<Driver, Object> drivers = new IdentityHashMap<Driver, Object>();
	private final static ReentrantLock lock = new ReentrantLock();
	private final static Condition cond = lock.newCondition();
	private final static Map<Manager, Thread> managers = new IdentityHashMap<Manager, Thread>();

	private enum State {
		RUNNING, CLOSING, CLOSED
	}

	private volatile static State state = State.CLOSED;
	static {
		try {
			Class.forName("limax.net.ServerManagerImpl");
		} catch (ClassNotFoundException e) {
		}
		try {
			Class.forName("limax.net.ClientManagerImpl");
		} catch (ClassNotFoundException e) {
		}
	}

	private Engine() {
	}

	static int checkLimitProtocolSize(int size) throws SizePolicyException {
		if (size > limitProtocolSize)
			throw new SizePolicyException("size = " + size
					+ " exceed limax.net.Engine.limitProtocolSize = "
					+ limitProtocolSize);
		return size;
	}

	public static void registerDriver(Driver driver) {
		lock.lock();
		try {
			Class<?> configClass = driver.getConfigClass();
			if (!configClass.isInterface())
				throw new RuntimeException("Config class "
						+ configClass.getName() + " is not interface");
			Class<?> listenerClass = driver.getListenerClass();
			if (!listenerClass.isInterface())
				throw new RuntimeException("Listener class "
						+ listenerClass.getName() + " is not interface");
			Driver old = testDriver(configClass, listenerClass);
			if (null != old)
				throw new RuntimeException(
						"Register conflicted. Config class ("
								+ configClass.getName() + "/"
								+ old.getConfigClass().getName()
								+ ") and Listener class ("
								+ listenerClass.getName() + "/"
								+ old.getListenerClass().getName() + ")");
			drivers.put(driver, dummy);
		} finally {
			lock.unlock();
		}
	}

	private static Set<Class<?>> interfaceSet(Class<?> interfaceClass) {
		Set<Class<?>> set = new HashSet<Class<?>>();
		for (Class<?> c : interfaceClass.getInterfaces())
			set.addAll(interfaceSet(c));
		set.add(interfaceClass);
		return set;
	}

	private static Driver testDriver(Class<?> configClass,
			Class<?> listenerClass) {
		for (final Driver driver : drivers.keySet())
			if (interfaceSet(configClass).contains(driver.getConfigClass())
					|| interfaceSet(listenerClass).contains(
							driver.getListenerClass())
					|| interfaceSet(driver.getConfigClass()).contains(
							configClass)
					|| interfaceSet(driver.getListenerClass()).contains(
							listenerClass))
				return driver;
		return null;
	}

	public static void open(int netProcessors, int protocolSchedulers,
			int applicationExecutors) throws Exception {
		open(1, netProcessors, protocolSchedulers, applicationExecutors);
	}

	public static void open(int nioCpus, int netProcessors,
			int protocolSchedulers, int applicationExecutors) throws Exception {
		if (netProcessors < 1 || protocolSchedulers < 1
				|| applicationExecutors < 1 || nioCpus < 1)
			throw new IllegalArgumentException("any engine parameter must > 0");
		lock.lock();
		try {
			if (state != State.CLOSED)
				throw new IllegalStateException("engine is not closed!");
			NetModel.initialize(new FixedCpuPoll(nioCpus), netProcessors);
			ConcurrentEnvironment env = ConcurrentEnvironment.getInstance();
			env.newThreadPool("limax.net.Engine.applicationExecutor", 0);
			applicationExecutor = env.newHashExecutor(
					"limax.net.Engine.applicationExecutor",
					applicationExecutors);
			protocolScheduler = env.newScheduledThreadPool(
					"limax.net.Engine.protocolScheduler", protocolSchedulers);
			protocolExecutor = env.newHashExecutor(
					"limax.net.Engine.protocolScheduler", protocolSchedulers);
			engineExecutor = env.newThreadPool("limax.net.Engine.executor", 0);
			state = State.RUNNING;
		} finally {
			lock.unlock();
		}
	}

	public static void close() {
		Collection<Manager> c;
		lock.lock();
		try {
			if (state != State.RUNNING)
				throw new IllegalStateException("engine is not running!");
			state = State.CLOSING;
			c = new ArrayList<Manager>(managers.keySet());
		} finally {
			lock.unlock();
		}
		try {
			for (Manager manager : c)
				manager.close();
			ConcurrentEnvironment.getInstance().shutdown(
					"limax.net.Engine.executor",
					"limax.net.Engine.protocolScheduler",
					"limax.net.Engine.applicationExecutor");
			NetModel.uninitialize();
		} finally {
			state = State.CLOSED;
		}
	}

	public static Manager add(Config config, Listener listener, Manager wrapper)
			throws Exception {
		lock.lock();
		try {
			if (state != State.RUNNING)
				throw new IllegalStateException("engine is not running!");
			for (Driver driver : drivers.keySet()) {
				Class<?> configClass = driver.getConfigClass();
				Class<?> listenerClass = driver.getListenerClass();
				if (configClass.isInstance(config)
						&& listenerClass.isInstance(listener)) {
					Manager manager = driver.newInstance(config, listener,
							wrapper);
					managers.put(manager, null);
					return manager;
				}
			}
			throw new InstantiationException(config.getClass().getName()
					+ " with " + listener.getClass().getName()
					+ " driver not registered");
		} finally {
			lock.unlock();
		}
	}

	public static Manager add(Config config, Listener listener)
			throws Exception {
		return add(config, listener, null);
	}

	public static boolean remove(final Manager manager) {
		lock.lock();
		try {
			if (managers.containsKey(manager)) {
				Thread thread = managers.get(manager);
				if (thread == null) {
					engineExecutor.execute(new Runnable() {
						@Override
						public void run() {
							lock.lock();
							try {
								managers.put(manager, Thread.currentThread());
							} finally {
								lock.unlock();
							}
							manager.close();
						}
					});
					while (managers.get(manager) == null)
						try {
							cond.await();
						} catch (InterruptedException e) {
						}
					managers.remove(manager);
				} else if (thread == Thread.currentThread()) {
					cond.signalAll();
					return false;
				}
			}
			return true;
		} finally {
			lock.unlock();
		}
	}

	public static boolean contains(Manager manager) {
		lock.lock();
		try {
			return managers.containsKey(manager);
		} finally {
			lock.unlock();
		}
	}

	public static ScheduledExecutorService getProtocolScheduler() {
		return protocolScheduler;
	}

	public static HashExecutor getProtocolExecutor() {
		return protocolExecutor;
	}

	public static HashExecutor getApplicationExecutor() {
		return applicationExecutor;
	}
}
