package limax.provider;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.IdentityHashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.LockSupport;

import limax.codec.CodecException;
import limax.codec.StringStream;
import limax.defines.VariantDefines;
import limax.net.Engine;
import limax.net.SizePolicyException;
import limax.provider.providerendpoint.SyncViewToClients;
import limax.util.Enable;
import limax.util.Resource;

abstract class ViewContextImpl implements ViewContext {
	private final static long accumulateDelay = Long.getLong(
			"limax.provider.ViewContextImpl.accumulateDelay", 10).longValue();
	private final Resource resources = Resource.createRoot();
	private final Map<Short, GlobalView> globalViews = new HashMap<>();
	private final List<ViewStub> sessionViewStubs = new ArrayList<>();
	private final Map<Short, ViewStub> temporaryViewStubs = new HashMap<>();
	private final Map<ViewStub, Constructor<? extends View>> constructors = new IdentityHashMap<>();
	private final Set<TemporaryView> temporaryViews = Collections
			.newSetFromMap(Collections
					.synchronizedMap(new IdentityHashMap<TemporaryView, Boolean>()));
	private final AtomicInteger viewCounter = new AtomicInteger();
	private final ProviderManagerConfig config;
	private final VariantDefines variantDefines;
	private final Map<String, Integer> dataDictionary;
	private final String scriptViewNames;
	private volatile Thread closingThread;

	ViewContextImpl(ProviderManagerConfig config, String viewManagerClassName)
			throws Exception {
		Map<String, Integer> dataDictionary = new LinkedHashMap<>();
		String scriptViewNames = null;
		VariantDefines variantDefines = null;
		this.config = config;
		final Class<?> cls = Class.forName(viewManagerClassName);
		Method method = cls.getDeclaredMethod("initialize", ViewContext.class);
		method.setAccessible(true);
		@SuppressWarnings("unchecked")
		Collection<ViewStub> stubs = (Collection<ViewStub>) method.invoke(null,
				this);

		try {
			method = cls.getDeclaredMethod("getDataDictionary");
			method.setAccessible(true);
			int i = 0;
			for (String var : ((String) method.invoke(null)).split(","))
				dataDictionary.put(var, i++);
			StringStream sstream = StringStream.create("M");
			stubs.forEach(stub -> {
				sstream.marshal(stub.getClassIndex()).append("L");
				String s[] = stub.getViewClass().getName().split("\\.");
				for (int j = 2; j < s.length; j++)
					sstream.marshal(dataDictionary.get(s[j]));
				sstream.append(":");
			});
			scriptViewNames = sstream.toString(":");
		} catch (NoSuchMethodException | SecurityException
				| IllegalAccessException | IllegalArgumentException
				| InvocationTargetException e) {
			scriptViewNames = null;
		}

		try {
			method = cls.getDeclaredMethod("getVariantDefines");
			method.setAccessible(true);
			variantDefines = (VariantDefines) method.invoke(null);
		} catch (NoSuchMethodException | SecurityException
				| IllegalAccessException | IllegalArgumentException
				| InvocationTargetException e) {
			variantDefines = null;
		}
		this.variantDefines = variantDefines;
		this.dataDictionary = Collections.unmodifiableMap(dataDictionary);
		this.scriptViewNames = scriptViewNames;

		Constructor<? extends View> constructor;
		for (ViewStub stub : stubs) {
			method = stub
					.getViewClass()
					.getSuperclass()
					.getDeclaredMethod("registerTableListener",
							resources.getClass());
			method.setAccessible(true);
			method.invoke(null, resources);
			switch (stub.getLifecycle()) {
			case global:
				constructor = stub.getViewClass().getDeclaredConstructor(
						ViewContext.class, ViewStub.class, Resource.class);
				constructor.setAccessible(true);
				this.globalViews.put(stub.getClassIndex(),
						(GlobalView) onOpened(constructor.newInstance(this,
								stub, Resource.createRoot())));
				break;
			case session:
				constructor = stub.getViewClass().getDeclaredConstructor(
						ViewContext.class, ViewStub.class, Resource.class,
						long.class);
				constructor.setAccessible(true);
				this.sessionViewStubs.add(stub);
				this.constructors.put(stub, constructor);
				break;
			case temporary:
				constructor = stub.getViewClass().getDeclaredConstructor(
						ViewContext.class, ViewStub.class, Resource.class);
				constructor.setAccessible(true);
				this.temporaryViewStubs.put(stub.getClassIndex(), stub);
				this.constructors.put(stub, constructor);
			}
		}
	}

	abstract void syncViewToClients(SyncViewToClients protocol)
			throws InstantiationException, ClassCastException,
			SizePolicyException, CodecException;

	abstract ViewSession getViewSession(long sessionid);

	void delaySchedule(Runnable task) {
		Engine.getProtocolScheduler().schedule(task, accumulateDelay,
				TimeUnit.MILLISECONDS);
	}

	void hashSchedule(Object key, Runnable task) {
		Engine.getApplicationExecutor().execute(key, task);
	}

	void close() {
		closingThread = Thread.currentThread();
		temporaryViews.forEach(view -> view.close());
		globalViews.values().forEach(view -> view.close());
		while (viewCounter.get() != 0)
			LockSupport.park(viewCounter);
		resources.close();
	}

	View onOpened(View view) {
		viewCounter.incrementAndGet();
		return view;
	}

	void onClosed(View view) {
		if (viewCounter.decrementAndGet() == 0)
			LockSupport.unpark(closingThread);
	}

	void onSessionLogin(long sessionid, ViewSession session) throws Exception {
		for (ViewStub stub : sessionViewStubs)
			session.add((SessionView) onOpened(constructors.get(stub)
					.newInstance(this, stub, Resource.createRoot(), sessionid)));
	}

	View findView(long sessionid, short classindex, int instanceindex) {
		ViewSession session = getViewSession(sessionid);
		if (session == null)
			return null;
		if (instanceindex != 0)
			return session.findTemporaryView(classindex, instanceindex);
		View view = session.findSessionView(classindex);
		if (null != view)
			return view;
		view = findGlobalView(classindex);
		if (null != view)
			return view;
		return session.findTemporaryView(classindex, instanceindex);
	}

	@Override
	public GlobalView findGlobalView(short classindex) {
		return globalViews.get(classindex);
	}

	@Override
	public SessionView findSessionView(long sessionid, short classindex) {
		final ViewSession session = getViewSession(sessionid);
		return session != null ? session.findSessionView(classindex) : null;
	}

	@Override
	public TemporaryView findTemporaryView(long sessionid, short classindex,
			int instanceindex) {
		final ViewSession session = getViewSession(sessionid);
		return session != null ? session.findTemporaryView(classindex,
				instanceindex) : null;
	}

	@Override
	public TemporaryView createTemporaryView(short classindex) {
		try {
			ViewStub stub = temporaryViewStubs.get(classindex);
			TemporaryView view = (TemporaryView) onOpened(constructors
					.get(stub).newInstance(this, stub, Resource.createRoot()));
			temporaryViews.add(view);
			return view;
		} catch (Exception e) {
			throw new RuntimeException("createTemporaryView with classindex = "
					+ classindex, e);
		}
	}

	@Override
	public void closeTemporaryView(TemporaryView view) {
		if (temporaryViews.remove(view))
			view.close();
	}

	@Override
	public short getProviderId() {
		return config.getProviderId();
	}

	boolean isScriptSupported() {
		return scriptViewNames != null;
	}

	boolean isScriptEnabled() {
		return isScriptSupported()
				&& Enable.False != config.getAllowUseScript();
	}

	boolean isVariantEnabled() {
		return isVariantSupported()
				&& Enable.False != config.getAllowUseVariant();
	}

	boolean isVariantSupported() {
		return null != variantDefines;
	}

	VariantDefines getVariantDefines() {
		return variantDefines;
	}

	@Override
	public Map<String, Integer> getDataDictionary() {
		return dataDictionary;
	}

	String getScriptDefines() {
		return StringStream.create()
				.marshal(String.join(",", dataDictionary.keySet()))
				.append(scriptViewNames).toString();
	}
}
