package limax.switcher;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import limax.codec.Octets;
import limax.defines.ErrorCodes;
import limax.defines.SessionType;
import limax.defines.VariantDefines;
import limax.net.Config;
import limax.net.Manager;
import limax.net.Protocol;
import limax.net.ServerListener;
import limax.net.ServerManagerConfig;
import limax.net.SizePolicyException;
import limax.net.Transport;
import limax.provider.DirectDispatcher;
import limax.provider.providerendpoint.SendControlToServer;
import limax.switcher.switcherauany.CheckProviderKey;
import limax.switcher.switcherprovider.Bind;
import limax.switcher.switcherprovider.BindResult;
import limax.switcher.switcherprovider.Dispatch;
import limax.switcher.switcherprovider.LinkBroken;
import limax.switcher.switcherprovider.UnBind;
import limax.util.Trace;

public final class ProviderListener implements ServerListener {

	private static final ProviderListener instance = new ProviderListener();

	public static ProviderListener getInstance() {
		return instance;
	}

	private ProviderListener() {
	}

	private volatile Manager manager;
	private int listenPort;

	private final Map<Short, Transport> pvidmap = new HashMap<>();

	private static class SessionObject implements ProviderArgs {
		private final short pvid;
		private final Map<Integer, Integer> pinfos;
		private final Set<Long> sids = new HashSet<>();
		private final VariantDefines variantDefines;
		private final String scriptDefines;
		private final int status;

		SessionObject(short pvid, Map<Integer, Integer> pinfos, int status,
				VariantDefines variantDefines, String scriptDefines) {
			this.pvid = pvid;
			this.pinfos = pinfos;
			this.status = status;
			this.variantDefines = variantDefines;
			this.scriptDefines = scriptDefines;
		}

		private boolean isStatusTrue(int mask) {
			return mask == (status & mask);
		}

		@Override
		public boolean isVarinatEnabled() {
			return isStatusTrue(Bind.PS_VARIANT_ENABLED);
		}

		@Override
		public boolean isVariantSupported() {
			return isStatusTrue(Bind.PS_VARIANT_SUPPORTED);
		}

		@Override
		public boolean isScriptEnabled() {
			return isStatusTrue(Bind.PS_SCRIPT_ENABLED);
		}

		@Override
		public boolean isScriptSupported() {
			return isStatusTrue(Bind.PS_SCRIPT_SUPPORTED);
		}

		@Override
		public boolean isStateless() {
			return isStatusTrue(Bind.PS_STATELESS);
		}

		@Override
		public VariantDefines getVariantDefines() {
			return variantDefines;
		}

		@Override
		public String getScriptDefines() {
			return scriptDefines;
		}

		@Override
		public short getProviderId() {
			return pvid;
		}

	}

	Manager getManager() {
		return manager;
	}

	@Override
	public void onManagerInitialized(Manager manager, Config config) {
		this.manager = manager;
		listenPort = ((InetSocketAddress) ((ServerManagerConfig) config)
				.getLocalAddress()).getPort();
	}

	int getListenPort() {
		return listenPort;
	}

	@Override
	public void onManagerUninitialized(Manager manager) {
	}

	@Override
	public void onTransportAdded(Transport transport) {
	}

	@Override
	public void onTransportRemoved(Transport transport) {
		doUnBind(transport);
	}

	private void closeTransport(Transport transport) {
		manager.close(transport);
	}

	ProviderArgs getProviderArgs(Transport transport) {
		return (SessionObject) transport.getSessionObject();
	}

	synchronized ProviderArgs getProviderArgs(short pvid) {
		Transport transport = pvidmap.get(pvid);
		return transport == null ? null : (SessionObject) transport
				.getSessionObject();
	}

	private final Map<Transport, CheckProviderKey> checkPending = new IdentityHashMap<>();

	public final synchronized void processResult(CheckProviderKey rpc) {
		final Bind bind = (Bind) rpc.getNote();
		final Transport transport = bind.getTransport();
		try {
			checkPending.remove(transport);
			final int errorcode = rpc.getResult().errorcodes;
			if (ErrorCodes.SUCCEED != errorcode) {
				new BindResult(bind.pvid, errorcode).send(transport);
				return;
			}
			if (pvidmap.containsKey(bind.pvid)) {
				if (Trace.isErrorEnabled())
					Trace.error("Transport " + transport
							+ " closed. Duplicate pvid = " + bind.pvid);
				new BindResult(bind.pvid, ErrorCodes.PROVIDER_DUPLICATE_ID)
						.send(transport);
				return;
			}
			transport.setSessionObject(new SessionObject(bind.pvid,
					bind.pinfos, bind.providerStatus, bind.variantdefines,
					bind.scriptdefines));
			pvidmap.put(bind.pvid, transport);
			new BindResult(bind.pvid, ErrorCodes.SUCCEED).send(transport);
			if (Trace.isErrorEnabled())
				Trace.info("provider bind " + bind.pvid);
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("provider send protocol exception", e);
			closeTransport(transport);
		}
		final DirectDispatcher.ProviderDispatchable dispatch = DirectDispatcher
				.getInstance().getProviderDispatchable(bind.pvid);
		if (null == dispatch)
			DirectDispatcher.getInstance().register(bind.pvid,
					createProviderDispatchable(transport));
		else
			dispatch.setSwitcherReceiveable(createSwitcherReceiveable());
	}

	public final synchronized void processTimeout(CheckProviderKey rpc) {
		try {
			rpc.send(AuanyClientListener.getInstance().getTransport());
		} catch (Exception e) {
		}
	}

	void flushCheckPending(Transport transport) {
		checkPending.values().forEach(rpc -> {
			try {
				rpc.send(transport);
			} catch (Exception e) {
			}
		});
	}

	void doUnBind(Transport transport) {
		final SessionObject so;
		synchronized (this) {
			checkPending.remove(transport);
			so = (SessionObject) transport.getSessionObject();
			if (so == null)
				return;
			pvidmap.remove(so.pvid);
			transport.setSessionObject(null);
		}
		DirectDispatcher.getInstance().unregister(so.pvid);
		so.sids.forEach(sid -> SwitcherListener.getInstance().kickSession(sid,
				ErrorCodes.SWITCHER_PROVIDER_UNBIND));
	}

	public final synchronized void process(Bind bind) {
		final CheckProviderKey rpc = new CheckProviderKey();
		rpc.getArgument().pvid = bind.pvid;
		rpc.getArgument().pvkey = bind.pvkey;
		rpc.setNote(bind);
		checkPending.put(bind.getTransport(), rpc);
		try {
			rpc.send(AuanyClientListener.getInstance().getTransport());
		} catch (Exception e) {
		}
	}

	public final void process(UnBind protocol) {
		doUnBind(protocol.getTransport());
	}

	private DirectDispatcher.ProviderDispatchable createProviderDispatchable(
			final Transport transport) {
		return new DirectDispatcher.ProviderDispatchable() {
			@Override
			public void dispatchSessionProtocol(long sid, int ptype,
					Octets pdata) {
				try {
					new Dispatch(sid, ptype, pdata).send(transport);
				} catch (Exception e) {
					if (Trace.isErrorEnabled())
						Trace.error(
								"Switcher send dispatch exception, kickSession sessionid = "
										+ sid, e);
					SwitcherListener.getInstance().kickSession(sid,
							ErrorCodes.SWITCHER_SEND_DISPATCH_EXCEPTION);
				}
			}

			@Override
			public void setSwitcherReceiveable(
					DirectDispatcher.SwitcherReceiveable recvable) {
				throw new UnsupportedOperationException();
			}

			@Override
			public void onViewControl(SendControlToServer protocol) {
				try {
					protocol.send(transport);
				} catch (Exception e) {
					if (Trace.isErrorEnabled())
						Trace.error("Switcher onViewControl protocol = "
								+ protocol, e);
				}
			}
		};
	}

	private DirectDispatcher.SwitcherReceiveable createSwitcherReceiveable() {
		return new DirectDispatcher.SwitcherReceiveable() {
			@Override
			public void switcherUnicast(long sid, int ptype, Octets data) {
				SwitcherListener.getInstance().doProviderUnicast(sid, ptype,
						data);
			}

			@Override
			public void switcherMulticast(Collection<Long> sids, int ptype,
					Octets data) {
				SwitcherListener.getInstance().doProviderMulticast(sids, ptype,
						data);
			}

			@Override
			public void switcherBroadcast(int ptype, Octets data) {
				SwitcherListener.getInstance().doProviderBroadcast(ptype, data);
			}

			@Override
			public void syncViewToClients(
					limax.provider.providerendpoint.SyncViewToClients p) {
				try {
					SwitcherListener.getInstance().process(p);
				} catch (Exception e) {
					if (Trace.isErrorEnabled())
						Trace.error("syncViewToClients protocol = " + p, e);
				}
			}
		};
	}

	final synchronized boolean checkProtocolSize(int type, int size)
			throws InstantiationException, SizePolicyException {
		Transport transport = pvidmap.get(DirectDispatcher.getProviderId(type));
		if (transport == null)
			return false;
		SessionObject so = (SessionObject) transport.getSessionObject();
		if (so == null)
			return false;
		Integer maxSize = so.pinfos.get(type);
		if (maxSize == null)
			throw new InstantiationException();
		if (maxSize < size)
			throw new SizePolicyException();
		return true;
	}

	final void broadcastLinkBroken(long sid, Collection<Short> pvids) {
		Collection<Transport> transports = new ArrayList<>(pvids.size());
		synchronized (this) {
			pvids.stream()
					.map(pvid -> pvidmap.get(pvid))
					.filter(Objects::nonNull)
					.forEach(
							transport -> {
								SessionObject so = (SessionObject) transport
										.getSessionObject();
								if (so != null && !so.isStateless()) {
									so.sids.remove(sid);
									transports.add(transport);
								}
							});
		}
		LinkBroken protocol = new LinkBroken(sid);
		transports
				.forEach(transport -> {
					try {
						protocol.send(transport);
					} catch (Exception e) {
						if (Trace.isErrorEnabled())
							Trace.error(manager + " send LinkBroken to "
									+ transport, e);
					}
				});
	}

	synchronized int checkProvidersReady(Map<Short, Byte> pvids) {
		for (Map.Entry<Short, Byte> e : pvids.entrySet()) {
			ProviderArgs args = getProviderArgs(e.getKey());
			if (null == args)
				return ErrorCodes.SWITCHER_LOST_PROVIDER;
			if (Main.isSessionUseVariant(e.getValue())) {
				if (!args.isVariantSupported())
					return ErrorCodes.PROVIDER_UNSUPPORTED_VARINAT;
				else if (!args.isVarinatEnabled())
					return ErrorCodes.PROVIDER_NOT_ALLOW_VARINAT;
			}
			if (Main.isSessionUseScript(e.getValue())) {
				if (!args.isScriptSupported())
					return ErrorCodes.PROVIDER_UNSUPPORTED_SCRIPT;
				else if (!args.isScriptEnabled())
					return ErrorCodes.PROVIDER_NOT_ALLOW_SCRIPT;
			}
			if (args.isStateless())
				pvids.put(e.getKey(),
						(byte) (e.getValue() | SessionType.ST_STATELESS));
		}
		return ErrorCodes.SUCCEED;
	}

	synchronized Collection<Long> getSessionIdsByProtocolType(int type) {
		Transport transport = pvidmap.get(DirectDispatcher.getProviderId(type));
		if (transport != null) {
			SessionObject so = (SessionObject) transport.getSessionObject();
			if (so != null)
				return new ArrayList<>(so.sids);
		}
		return Collections.emptyList();
	}

	synchronized void registerClient(long sid, Collection<Short> pvids) {
		pvids.stream().map(pvid -> pvidmap.get(pvid)).filter(Objects::nonNull)
				.map(t -> (SessionObject) t.getSessionObject())
				.filter(Objects::nonNull).filter(so -> !so.isStateless())
				.forEach(so -> so.sids.add(sid));
	}

	synchronized void unregisterClient(long sid, Collection<Short> pvids) {
		pvids.stream().map(pvid -> pvidmap.get(pvid)).filter(Objects::nonNull)
				.map(t -> (SessionObject) t.getSessionObject())
				.filter(Objects::nonNull).filter(so -> !so.isStateless())
				.forEach(so -> so.sids.remove(sid));
	}

	synchronized void multicastOnlineAnnounce(Collection<Short> pvids,
			Protocol protocol) {
		pvids.stream()
				.map(pvid -> pvidmap.get(pvid))
				.filter(Objects::nonNull)
				.filter(transport -> !((SessionObject) transport
						.getSessionObject()).isStateless())
				.forEach(
						transport -> {
							try {
								protocol.send(transport);
							} catch (Exception e) {
								if (Trace.isErrorEnabled())
									Trace.error(manager + " multicast "
											+ protocol + " to " + transport, e);
							}
						});
	}
}
