package limax.switcher;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.RejectedExecutionException;

import limax.codec.Codec;
import limax.codec.CodecException;
import limax.codec.HmacMD5;
import limax.codec.Octets;
import limax.codec.OctetsStream;
import limax.codec.StringStream;
import limax.defines.ErrorCodes;
import limax.defines.ErrorSource;
import limax.defines.SessionFlags;
import limax.defines.SessionType;
import limax.net.Config;
import limax.net.Manager;
import limax.net.Protocol;
import limax.net.ServerListener;
import limax.net.ServerManager;
import limax.net.SizePolicyException;
import limax.net.StateTransport;
import limax.net.Transport;
import limax.net.UnknownProtocolHandler;
import limax.net.WebSocketProtocol;
import limax.net.WebSocketTransport;
import limax.net.io.WebSocketAddress;
import limax.provider.DirectDispatcher;
import limax.providerendpoint.ViewMemberData;
import limax.providerendpoint.ViewVariableData;
import limax.switcher.providerendpoint.SendControlToServer;
import limax.switcher.states.SwitcherServer;
import limax.switcher.switcherauany.SessionAuthByToken;
import limax.switcher.switcherendpoint.CHandShake;
import limax.switcher.switcherendpoint.PingAndKeepAlive;
import limax.switcher.switcherendpoint.PortForward;
import limax.switcher.switcherendpoint.SHandShake;
import limax.switcher.switcherendpoint.SessionKick;
import limax.switcher.switcherendpoint.SessionLoginByToken;
import limax.switcher.switcherprovider.Broadcast;
import limax.switcher.switcherprovider.Kick;
import limax.switcher.switcherprovider.Multicast;
import limax.switcher.switcherprovider.NotifyDuplicateSession;
import limax.switcher.switcherprovider.Unicast;
import limax.switcherauany.AuanyAuthArg;
import limax.switcherauany.AuanyAuthRes;
import limax.util.ConcurrentEnvironment;
import limax.util.Helper;
import limax.util.Trace;

public final class SwitcherListener implements ServerListener {
	private final static long handShakeTimeout = Long.getLong(
			"limax.switcher.SwitcherListener.handShakeTimeout", 1000);
	private final static long sessionLoginTimeout = Long.getLong(
			"limax.switcher.SwitcherListener.sessionLoginTimeout", 20000);
	private final static long keepAliveTimeout = Long.getLong(
			"limax.switcher.SwitcherListener.keepAliveTimeout", 60000);
	private final static long pingProtect = Long.getLong(
			"limax.switcher.SwitcherListener.pingProtect", 30000);
	private final List<ServerManager> managers = new ArrayList<>();

	private static class SessionObject {
		final long sessionid;
		final long flags;
		final Map<Short, Byte> pvids;
		private Set<Short> _readyset = new HashSet<>();
		private Collection<Object> _cache = new ArrayList<>();
		private long pingtime = 0;

		SessionObject(long sessionid, long flags, Map<Short, Byte> pvids) {
			this.sessionid = sessionid;
			this.flags = flags;
			this.pvids = pvids;

			pvids.entrySet()
					.stream()
					.filter(e -> SessionType.ST_STATELESS == (SessionType.ST_STATELESS & e
							.getValue()))
					.forEach(e -> _readyset.add(e.getKey()));
		}

		@Override
		public String toString() {
			return "SessionObject sessionid=" + sessionid;
		}

		public boolean isCanFlowControl() {
			return SessionFlags.FLAG_CAN_FLOW_CONTROL == (SessionFlags.FLAG_CAN_FLOW_CONTROL & flags);
		}

		public boolean isHasPermission() {
			return SessionFlags.FLAG_HAS_PERMISSION == (SessionFlags.FLAG_HAS_PERMISSION & flags);
		}

		private boolean _isReady() {
			return null == _readyset || _readyset.size() == pvids.size();
		}

		static enum ReadyStatus {
			RS_READY_BEFORE, RS_UNREADY, RS_READY_NOW,
		}

		synchronized boolean cacheIfUnready(Object p) {
			if (_isReady())
				return false;
			_cache.add(p);
			return true;
		}

		synchronized ReadyStatus providerReady(short pvid) {
			if (_isReady())
				return ReadyStatus.RS_READY_BEFORE;
			_readyset.add(pvid);
			return _isReady() ? ReadyStatus.RS_READY_NOW
					: ReadyStatus.RS_UNREADY;
		}

		synchronized Collection<Object> ready() {
			Collection<Object> result = _cache;
			_readyset = null;
			_cache = null;
			return result;
		}

		boolean isPingFlood() {
			long current = System.currentTimeMillis();
			if (current - pingtime < pingProtect)
				return true;
			pingtime = current;
			return false;
		}
	}

	private SwitcherListener() {
	}

	private final static SwitcherListener instance = new SwitcherListener();

	public static SwitcherListener getInstance() {
		return instance;
	}

	final UnknownProtocolHandler unknownProtocolHandler = new UnknownProtocolHandler() {
		@Override
		public void check(int type, int size, Transport transport)
				throws InstantiationException, SizePolicyException {
			SessionObject so = (SessionObject) transport.getSessionObject();
			int errcode = ErrorCodes.SUCCEED;
			if (so.pvids.containsKey(DirectDispatcher.getProviderId(type))) {
				if (!ProviderListener.getInstance().checkProtocolSize(type,
						size))
					errcode = ErrorCodes.SWITCHER_LOST_PROVIDER;
			} else
				errcode = ErrorCodes.SWITCHER_WRONG_PROVIDER;
			if (errcode != ErrorCodes.SUCCEED)
				SwitcherListener.getInstance().kickSession(transport, errcode);
		}

		@Override
		public void dispatch(int type, Octets data, Transport from)
				throws CodecException {
			DirectDispatcher
					.getInstance()
					.getProviderDispatchableByProtocolType(type)
					.dispatchSessionProtocol(
							((SessionObject) from.getSessionObject()).sessionid,
							type, data);
			((StateTransport) from).resetAlarm(keepAliveTimeout);
		}
	};

	private final ConcurrentHashMap<Long, Transport> session2iomap = new ConcurrentHashMap<>();
	private ExecutorService multiSendExecutor;

	@Override
	public void onManagerInitialized(Manager manager, Config config) {
		synchronized (managers) {
			if (managers.size() == 0)
				multiSendExecutor = ConcurrentEnvironment.getInstance()
						.newFixedThreadPool("limax.switcher.multiSender", 1);
			managers.add((ServerManager) manager);
		}
	}

	@Override
	public void onManagerUninitialized(Manager manager) {
		synchronized (managers) {
			managers.remove(manager);
			if (managers.size() == 0)
				ConcurrentEnvironment.getInstance().shutdown(
						"limax.switcher.multiSender");
		}
	}

	@Override
	public void onTransportRemoved(Transport transport)
			throws InstantiationException, SizePolicyException, CodecException {
		if (Trace.isDebugEnabled())
			Trace.debug("SwitcherListener onTransportRemoved transport = "
					+ transport + " " + transport.getCloseReason());
		final SessionObject so = (SessionObject) transport.getSessionObject();
		if (null == so)
			return;
		transport.setSessionObject(null);
		session2iomap.remove(so.sessionid);
		ProviderListener.getInstance().unregisterClient(so.sessionid,
				so.pvids.keySet());
		ProviderListener.getInstance().broadcastLinkBroken(so.sessionid,
				so.pvids.keySet());
	}

	private static byte[] packAddress(SocketAddress peeraddress,
			SocketAddress reportaddress) throws IOException {
		try (ByteArrayOutputStream os = new ByteArrayOutputStream();
				ObjectOutputStream oos = new ObjectOutputStream(os)) {
			oos.writeObject(peeraddress);
			oos.writeObject(reportaddress);
			return os.toByteArray();
		}
	}

	@Override
	public void onTransportAdded(Transport transport) throws Exception {
		if (Trace.isDebugEnabled())
			Trace.debug("SwitcherListener onTransportAdded transport = "
					+ transport);
		if (transport instanceof WebSocketTransport) {
			((WebSocketTransport) transport).resetAlarm(sessionLoginTimeout);
			final WebSocketAddress wsa = (WebSocketAddress) transport
					.getPeerAddress();
			final AuanyAuthArg arg = new AuanyAuthArg();
			final String query = URLDecoder.decode(wsa.getRequestURI()
					.getQuery().trim(), "UTF-8");
			final String[] kvs = query.split("&");
			for (final String kv : kvs) {
				final String[] pair = kv.trim().split("=");
				if (2 != pair.length)
					continue;
				final String k = pair[0].trim();
				final String v = pair[1].trim();
				switch (k) {
				case "username":
					arg.username = v;
					break;
				case "token":
					arg.token = v;
					break;
				case "platflag":
					arg.platflag = v;
					break;
				case "pvids":
					for (String sid : v.split(","))
						arg.pvids
								.put(Short.parseShort(sid.trim()),
										(byte) (SessionType.ST_SCRIPT | SessionType.ST_WEB_SOCKET));

				}
			}
			int errorcode = ProviderListener.getInstance().checkProvidersReady(
					arg.pvids);
			if (ErrorCodes.SUCCEED != errorcode) {
				new WebSocketProtocol(StringStream.create()
						.marshal(ErrorSource.LIMAX).marshal(errorcode)
						.toString()) {
					@Override
					public void process() throws Exception {
					}
				}.send(transport);
				close(transport);
				return;
			}
			arg.clientaddress.replace(packAddress(((WebSocketAddress) transport
					.getPeerAddress()).getSocketAddress(),
					new InetSocketAddress(0xffff)));
			arg.sockettype = AuanyAuthArg.ST_WEBSOCKET;
			final SessionAuthByToken auth = new SessionAuthByToken(arg);
			auth.setNote(transport);
			auth.send(AuanyClientListener.getInstance().getTransport());
		} else {
			((StateTransport) transport).resetAlarm(handShakeTimeout);
		}
	}

	private volatile boolean s2cneedcompress = true;
	private volatile boolean c2sneedcompress = true;
	private volatile Set<Integer> dhGroups;

	void setNeedCompress(boolean s2c, boolean c2s) {
		s2cneedcompress = s2c;
		c2sneedcompress = c2s;
	}

	void setDHGroups(Set<Integer> dhGroups) {
		this.dhGroups = dhGroups;
	}

	private void close(Transport transport) {
		transport.getManager().close(transport);
	}

	public final void process(PingAndKeepAlive p) throws Exception {
		StateTransport transport = (StateTransport) p.getTransport();
		SessionObject so = (SessionObject) transport.getSessionObject();
		if (so == null) {
			p.send(transport);
			close(transport);
		} else if (so.isPingFlood()) {
			if (Trace.isInfoEnabled())
				Trace.info("transport closed! " + transport
						+ " isPingFlood = true");
			close(transport);
		} else {
			p.send(transport);
			transport.resetAlarm(keepAliveTimeout);
		}
	}

	public final void process(PortForward p) throws Exception {
		StateTransport transport = (StateTransport) p.getTransport();
		SessionObject so = (SessionObject) transport.getSessionObject();
		transport.resetAlarm(keepAliveTimeout);
		if (so.isCanFlowControl()) {
			PortForwardManager.getInstance().onPortForward(p);
			return;
		}
		final PortForward protocol = new PortForward();
		protocol.command = PortForward.eClose;
		protocol.portsid = p.portsid;
		protocol.code = PortForward.eCloseNoAuthority;
		protocol.send(transport);
	}

	public final void process(CHandShake protocol) throws Exception {
		StateTransport transport = (StateTransport) protocol.getTransport();
		int group = protocol.dh_group;
		if (!dhGroups.contains(group)) {
			kickSession(transport, ErrorCodes.SWITCHER_DHGROUP_NOTSUPPRTED);
			return;
		}
		transport.resetAlarm(sessionLoginTimeout);
		BigInteger data = new BigInteger(protocol.dh_data.getBytes());
		BigInteger rand = Helper.makeDHRandom();
		byte[] material = Helper.computeDHKey(group, data, rand).toByteArray();
		byte[] key = ((InetSocketAddress) transport.getLocalAddress())
				.getAddress().getAddress();
		int half = material.length / 2;
		HmacMD5 mac = new HmacMD5(Codec.Null(), key, 0, key.length);
		mac.update(material, 0, half);
		transport.setInputSecurityCodec(mac.digest(), c2sneedcompress);
		new SHandShake(Octets.wrap(Helper.generateDHResponse(group, rand)
				.toByteArray()), c2sneedcompress, s2cneedcompress)
				.send(transport);
		mac = new HmacMD5(Codec.Null(), key, 0, key.length);
		mac.update(material, half, material.length - half);
		transport.setOutputSecurityCodec(mac.digest(), s2cneedcompress);
		transport.setState(SwitcherServer.EndpointSessionLogin);
	}

	public final void process(SessionLoginByToken p) throws Exception {
		final Transport transport = p.getTransport();
		final int errorcode = ProviderListener.getInstance()
				.checkProvidersReady(p.pvids);
		if (ErrorCodes.SUCCEED != errorcode) {
			kickSession(transport, errorcode);
			return;
		}
		final SessionAuthByToken auth = new SessionAuthByToken(
				new AuanyAuthArg(p.username, p.token, p.platflag, p.pvids,
						new Octets(packAddress(
								transport.getPeerAddress(),
								new InetSocketAddress(InetAddress
										.getByAddress(p.report_ip.getBytes()),
										p.report_port & 0xffff))),
						(byte) AuanyAuthArg.ST_SOCKET));
		auth.setNote(transport);
		auth.send(AuanyClientListener.getInstance().getTransport());
	}

	public final void processResult(SessionAuthByToken auth) {
		final AuanyAuthArg arg = auth.getArgument();
		final AuanyAuthRes res = auth.getResult();
		final Transport transport = (Transport) auth.getNote();
		final SessionObject so = new SessionObject(res.sessionid, res.flags,
				arg.pvids);
		final boolean loginsucceed = ErrorSource.LIMAX == res.errorSource
				&& ErrorCodes.SUCCEED == res.errorCode;
		boolean haslinknow = false;
		if (loginsucceed) {
			haslinknow = session2iomap.putIfAbsent(res.sessionid, transport) != null;
			if (haslinknow)
				res.errorCode = ErrorCodes.SWITCHER_SESSION_LOGINED;
		}
		try {
			if (loginsucceed && !haslinknow) {
				final limax.switcher.switcherprovider.OnlineAnnounce protocol = new limax.switcher.switcherprovider.OnlineAnnounce();
				protocol.sid = res.sessionid;
				protocol.clientaddress.swap(arg.clientaddress);
				protocol.flags = res.flags;
				if (so.isHasPermission())
					protocol.permissions.addAll(res.permissions);
				protocol.platflag = arg.platflag;
				protocol.sessiontype.putAll(so.pvids);
				ProviderListener.getInstance().multicastOnlineAnnounce(
						so.pvids.keySet(), protocol);
				ProviderListener.getInstance().registerClient(so.sessionid,
						so.pvids.keySet());
				transport.setSessionObject(so);
				if (transport instanceof StateTransport) {
					((StateTransport) transport)
							.setState(SwitcherServer.EndpointClient);
					((StateTransport) transport).resetAlarm(keepAliveTimeout);
				} else {
					((WebSocketTransport) transport)
							.resetAlarm(keepAliveTimeout);
				}
			} else {
				if (haslinknow)
					ProviderListener.getInstance().multicastOnlineAnnounce(
							so.pvids.keySet(),
							new NotifyDuplicateSession(res.sessionid));
				if (transport instanceof StateTransport) {
					final limax.switcher.switcherendpoint.OnlineAnnounce protocol = new limax.switcher.switcherendpoint.OnlineAnnounce();
					protocol.errorSource = res.errorSource;
					protocol.errorCode = res.errorCode;
					protocol.send(transport);
				} else {
					new WebSocketProtocol(StringStream.create()
							.marshal(res.errorSource).marshal(res.errorCode)
							.toString()) {
						@Override
						public void process() throws Exception {
						}
					}.send(transport);
				}
				close(transport);
			}
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("Transport " + transport
						+ " closed. SwitcherManager.onSessionloginDone", e);
			close(transport);
		}
	}

	public final void processTimeout(SessionAuthByToken auth) {
		final Transport transport = (Transport) auth.getNote();
		try {
			if (transport instanceof StateTransport) {
				final limax.switcher.switcherendpoint.OnlineAnnounce protocol = new limax.switcher.switcherendpoint.OnlineAnnounce();
				protocol.errorSource = ErrorSource.LIMAX;
				protocol.errorCode = ErrorCodes.SWITCHER_AUANY_TIMEOUT;
				protocol.send(transport);
			} else {
				new WebSocketProtocol(StringStream.create()
						.marshal(ErrorSource.LIMAX)
						.marshal(ErrorCodes.SWITCHER_AUANY_TIMEOUT).toString()) {
					@Override
					public void process() throws Exception {
					}
				}.send(transport);
			}
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("SwitcherManager.onSessionloginTimeout", e);
		} finally {
			close(transport);
		}
	}

	public final void process(limax.switcher.switcherendpoint.OnlineAnnounce p) {
		ProviderArgs args = ProviderListener.getInstance().getProviderArgs(
				p.getTransport());
		final Transport transport = session2iomap.get(p.sid);
		final SessionObject so = (SessionObject) transport.getSessionObject();
		switch (so.providerReady(args.getProviderId())) {
		case RS_READY_BEFORE:
			if (Trace.isErrorEnabled())
				Trace.error("session is ready bu recv OnlineAnnounce from provider!");
			return;
		case RS_UNREADY:
			return;
		case RS_READY_NOW:
			break;
		}
		try {
			if (transport instanceof WebSocketTransport) {
				StringStream sstream = StringStream.create()
						.marshal(p.errorSource).marshal(p.errorCode)
						.marshal(p.sid).marshal(p.flags).append("L");
				so.pvids.keySet().forEach(
						pvid -> sstream.marshal(pvid).append(
								ProviderListener.getInstance()
										.getProviderArgs(pvid)
										.getScriptDefines()));
				final String str = sstream.toString(":");
				new WebSocketProtocol(str) {
					@Override
					public void process() throws Exception {
					}
				}.send(transport);
				for (Object o : so.ready())
					((WebSocketProtocol) o).send(transport);
			} else {
				Set<Short> scriptpvids = new HashSet<>();
				so.pvids.forEach((k, v) -> {
					if (!Main.isSessionProtocolOnly(v)) {
						ProviderArgs a = ProviderListener.getInstance()
								.getProviderArgs(k);
						if (Main.isSessionUseVariant(v))
							p.variantdefines.put(a.getProviderId(),
									a.getVariantDefines());
						if (Main.isSessionUseScript(v))
							scriptpvids.add(a.getProviderId());
					}
				});
				if (!scriptpvids.isEmpty()) {
					StringStream sstream = StringStream.create()
							.marshal(p.errorSource).marshal(p.errorCode)
							.marshal(p.sid).marshal(p.flags).append("L");
					scriptpvids.forEach(pvid -> sstream.marshal(pvid).append(
							ProviderListener.getInstance()
									.getProviderArgs(pvid).getScriptDefines()));
					p.scriptdefines = sstream.toString(":");
				}
				p.send(transport);
				for (Object o : so.ready())
					((Protocol) o).send(transport);
			}
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("switcher send protocol exception", e);
			close(transport);
		}
	}

	public final void process(SendControlToServer p) {
		short providerid;
		short classindex;
		int instanceindex;
		byte controlindex;
		String message;
		if (p.providerid == -1) {
			String s = p.stringdata;
			providerid = Short.parseShort(s.substring(0, s.indexOf(',')),
					Character.MAX_RADIX);
			int p0 = s.indexOf(',') + 1;
			int p1 = s.indexOf(',', p0);
			classindex = Short.parseShort(s.substring(p0, p1),
					Character.MAX_RADIX);
			int p2 = s.indexOf(':', ++p1);
			instanceindex = Integer.parseInt(s.substring(p1, p2),
					Character.MAX_RADIX);
			controlindex = -1;
			message = s.substring(++p2);
		} else {
			providerid = p.providerid;
			classindex = p.classindex;
			instanceindex = p.instanceindex;
			controlindex = p.controlindex;
			message = p.stringdata;
		}
		StateTransport transport = (StateTransport) p.getTransport();
		SessionObject so = (SessionObject) transport.getSessionObject();
		if (!so.pvids.containsKey(providerid)) {
			kickSession(transport, ErrorCodes.SWITCHER_WRONG_PROVIDER);
			return;
		}
		DirectDispatcher.ProviderDispatchable dispatchable = DirectDispatcher
				.getInstance().getProviderDispatchable(providerid);
		if (dispatchable != null) {
			dispatchable
					.onViewControl(new limax.provider.providerendpoint.SendControlToServer(
							providerid, so.sessionid, classindex,
							instanceindex, controlindex, p.controlparameter,
							message));
			transport.resetAlarm(keepAliveTimeout);
		} else {
			kickSession(transport, ErrorCodes.SWITCHER_PROVIDER_UNBIND);
		}
	}

	public void process(WebSocketProtocol webSocketProtocol) {
		WebSocketTransport transport = webSocketProtocol.getTransport();
		String s = webSocketProtocol.getText();
		if (s.equals(" ")) { // cocos2dx-js websocket donot support empty string
			SessionObject so = (SessionObject) transport.getSessionObject();
			if (so == null || so.isPingFlood())
				close(transport);
			else
				transport.resetAlarm(keepAliveTimeout);
		} else {
			DirectDispatcher.ProviderDispatchable dispatchable = DirectDispatcher
					.getInstance().getProviderDispatchable(
							Short.parseShort(s.substring(0, s.indexOf(',')),
									Character.MAX_RADIX));
			if (dispatchable != null) {
				dispatchable.dispatchSessionProtocol(
						((SessionObject) webSocketProtocol.getTransport()
								.getSessionObject()).sessionid, -1,
						new OctetsStream().marshal(webSocketProtocol));
				transport.resetAlarm(keepAliveTimeout);
			} else {
				kickSession(transport, ErrorCodes.SWITCHER_PROVIDER_UNBIND);
			}
		}
	}

	private static class ForwardSyncViewToClients {

		final private Data data;
		final private short pvid;
		private Protocol protocol = null;
		private Protocol script = null;

		private ForwardSyncViewToClients(Data data, short pvid) {
			this.data = data;
			this.pvid = pvid;
		}

		public Protocol getProtocol() {
			if (null == protocol)
				protocol = data.createProtocol();
			return protocol;
		}

		public Protocol getScript() {
			if (null == script)
				script = data.createScript();
			return script;
		}

		public Protocol getFull() {
			return data.getFull();
		}

		public String getScriptString() {
			return data.getScriptString();
		}

		public short getProviderId() {
			return pvid;
		}

		private static interface Data {

			Protocol createProtocol();

			Protocol createScript();

			Protocol getFull();

			String getScriptString();
		}

		public static ForwardSyncViewToClients create(
				final limax.switcher.providerendpoint.SyncViewToClients p) {
			return new ForwardSyncViewToClients(new Data() {

				@Override
				public Protocol createProtocol() {
					return new limax.switcher.providerendpoint.SyncViewToClients(
							p.providerid, p.sessionids, p.classindex,
							p.instanceindex, p.synctype, p.vardatas, p.members,
							"");
				}

				@Override
				public Protocol createScript() {
					return new limax.switcher.providerendpoint.SyncViewToClients(
							p.providerid, p.sessionids, p.classindex,
							p.instanceindex, p.synctype,
							new ArrayList<ViewVariableData>(),
							new ArrayList<ViewMemberData>(), p.stringdata);
				}

				@Override
				public Protocol getFull() {
					return p;
				}

				@Override
				public String getScriptString() {
					return p.stringdata;
				}
			}, p.providerid);
		}

		public static ForwardSyncViewToClients create(
				final limax.provider.providerendpoint.SyncViewToClients p) {
			return new ForwardSyncViewToClients(new Data() {

				@Override
				public Protocol createProtocol() {
					return new limax.switcher.providerendpoint.SyncViewToClients(
							p.providerid, p.sessionids, p.classindex,
							p.instanceindex, p.synctype, p.vardatas, p.members,
							"");
				}

				@Override
				public Protocol createScript() {
					return new limax.switcher.providerendpoint.SyncViewToClients(
							p.providerid, p.sessionids, p.classindex,
							p.instanceindex, p.synctype,
							new ArrayList<ViewVariableData>(),
							new ArrayList<ViewMemberData>(), p.stringdata);
				}

				@Override
				public Protocol getFull() {
					return p;
				}

				@Override
				public String getScriptString() {
					return p.stringdata;
				}

			}, p.providerid);
		}
	}

	private interface forwardSyncViewToClientsOperator {

		Object getProtocol();

		void doSend() throws InstantiationException, ClassCastException,
				SizePolicyException, CodecException;
	}

	private void forwardSyncViewToClients(Collection<Long> sessionids,
			ForwardSyncViewToClients fscv) {
		for (Long sid : sessionids) {
			Transport transport = session2iomap.get(sid);
			if (null == transport)
				continue;
			SessionObject so = (SessionObject) transport.getSessionObject();
			final forwardSyncViewToClientsOperator operator;
			if (transport instanceof WebSocketTransport) {
				WebSocketProtocol protocol = new WebSocketProtocol(
						fscv.getScriptString()) {
					@Override
					public void process() throws Exception {
					}
				};
				operator = new forwardSyncViewToClientsOperator() {
					@Override
					public Object getProtocol() {
						return protocol;
					}

					@Override
					public void doSend() throws ClassCastException,
							CodecException {
						protocol.send(transport);
					}
				};
			} else {
				final Protocol protocol;
				final byte st = so.pvids.get(fscv.getProviderId());
				if (Main.isSessionUseViewProtocol(st)
						&& Main.isSessionUseScript(st))
					protocol = fscv.getFull();
				else if (Main.isSessionUseScript(st))
					protocol = fscv.getScript();
				else
					protocol = fscv.getProtocol();
				operator = new forwardSyncViewToClientsOperator() {
					@Override
					public Object getProtocol() {
						return protocol;
					}

					@Override
					public void doSend() throws InstantiationException,
							ClassCastException, SizePolicyException,
							CodecException {
						protocol.send(transport);
					}
				};
			}
			if (so.cacheIfUnready(operator.getProtocol()))
				continue;
			try {
				operator.doSend();
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error(transport
							+ " SwitcherManager.doSyncViewToClients", e);
				kickSession(transport,
						ErrorCodes.SWITCHER_SEND_TO_ENDPOINT_EXCEPTION);
			}
		}
	}

	private final static ArrayList<Long> nullList = new ArrayList<>();

	public void process(limax.provider.providerendpoint.SyncViewToClients p) {
		final Collection<Long> sessionids = p.sessionids;
		p.sessionids = nullList;
		forwardSyncViewToClients(sessionids, ForwardSyncViewToClients.create(p));
	}

	public void process(limax.switcher.providerendpoint.SyncViewToClients p) {
		final Collection<Long> sessionids = p.sessionids;
		p.sessionids = nullList;
		forwardSyncViewToClients(sessionids, ForwardSyncViewToClients.create(p));
	}

	void kickSession(long sid, int error) {
		Transport transport = session2iomap.get(sid);
		if (null != transport)
			kickSession(transport, error);
	}

	void kickSession(Transport transport, int error) {
		if (Trace.isInfoEnabled())
			Trace.info("switcher kick session " + transport + " " + error);
		try {
			if (transport instanceof WebSocketTransport)
				new WebSocketProtocol("" + error) {
					@Override
					public void process() throws Exception {
					}
				}.send(transport);
			else
				new SessionKick(error).send(transport);
		} catch (Exception e) {
			if (Trace.isErrorEnabled())
				Trace.error("SwitcherManager.kickSession send SessionKick", e);
		} finally {
			close(transport);
		}
	}

	private void doMultiSend(Collection<Long> sids, int type, Octets data) {
		multiSendExecutor.execute(() -> {
			sids.stream()
					.map(sid -> session2iomap.get(sid))
					.filter(Objects::nonNull)
					.forEach(
							transport -> {
								try {
									unknownProtocolHandler.send(type, data,
											transport);
								} catch (Exception e) {
									if (Trace.isErrorEnabled())
										Trace.error("multicast type = " + type
												+ " to " + transport, e);
								}
							});
		});
	}

	void doProviderUnicast(long sid, int ptype, Octets data) {
		Transport transport = session2iomap.get(sid);
		if (null != transport) {
			try {
				unknownProtocolHandler.send(ptype, data, transport);
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error(transport
							+ " closed. SwitcherManager.doProviderSend", e);
				kickSession(transport,
						ErrorCodes.SWITCHER_SEND_TO_ENDPOINT_EXCEPTION);
			}
		}
	}

	void doProviderMulticast(Collection<Long> sids, int type, Octets data) {
		try {
			if (!sids.isEmpty())
				doMultiSend(sids, type, data);
		} catch (RejectedExecutionException e) {
		}
	}

	void doProviderBroadcast(int ptype, Octets data) {
		doProviderMulticast(ProviderListener.getInstance()
				.getSessionIdsByProtocolType(ptype), ptype, data);
	}

	public void process(Kick p) {
		kickSession(p.sid, p.error);
	}

	public void process(Unicast p) {
		doProviderUnicast(p.sid, p.ptype, p.pdata);
	}

	public void process(Multicast p) {
		doProviderMulticast(p.sids, p.ptype, p.pdata);
	}

	public void process(Broadcast p) {
		doProviderBroadcast(p.ptype, p.pdata);
	}

}
